# ----------------------------------------------------------------------
# |  State-Manager
# ----------------------------------------------------------------------
from montyalex.console_tools import richconsole
from montyalex.fs_tools import cancel
from montyalex.object_tools import (
    Key,
    Value,
    KeyValueError,
    JSONError,
    RangeError,
)
from montyalex.typing_tools import Any, Literal
from montyalex.uo_tools import yaml

print = richconsole.print


# ----------------------------------------------------------------------
# |  State
# ----------------------------------------------------------------------
class State:
    """Handles state management using a YAML-based configuration system

    Attributes:
        multiplier (int): The multiplier for state modification.
        limit (int or tuple[int, int], optional): The limit for state range.
        ilimit (int or tuple[int, int], optional): The inclusive limit for state range.
        file_name (str): The name of the YAML file.
        file_path (str): The path to the YAML file.
        full_path (str): The full path to the YAML file.
        loc (str): The directory location of the YAML file.
        yamlchange: Method to change YAML content.
        yamlexists: Method to check if the YAML file exists.
        yamlmodel: Method to read or write YAML content.
        yamlremove: Method to remove the YAML file.
    """

    def __init__(
        self,
        file_name: str,
        *,
        multiplier: int = 1,
        limit: int | tuple[int, int] = None,
        ilimit: int | tuple[int, int] = None,
        loc: str = ".mtax",
    ) -> None:
        __yaml = yaml(directory=loc, filename=file_name)
        self.multiplier: int = multiplier
        self.limit: int | tuple[int, int] = limit
        self.ilimit: int | tuple[int, int] = ilimit
        self.file_name: str = file_name
        self.file_path: str = f"{file_name}.{__yaml.modelname}"
        self.full_path: str = __yaml.modelpath
        self.loc: str = loc
        self.yamlchange = __yaml.change
        self.yamlexists = __yaml.exists
        self.yamlmodel = __yaml.model
        self.yamlremove = __yaml.remove

    def get(self, key: Key):
        key_value = self.get_value(key)
        if key_value is not None:
            state: dict[Key, Any] = {key: key_value}
            print(f"Key Found {state}")
        else:
            raise KeyValueError("No value associated")

    def get_value(self, key: Key):
        try:
            key_value = self.yamlmodel(read=True)[key]
            if key_value is not None:
                return key_value
            raise KeyValueError(
                f"Null assignment for {key!r}; no value associated with active state"
            )
        except (KeyError, TypeError, JSONError):
            print(
                f"KeyError: {key!r} can not be found in file; no value was assigned"
            )
            return None

    def delete(
        self,
        key: Key = None,
        value: Value = None,
        *,
        fulloc: bool = False,
    ):
        if key is None and value is None:
            path = self.file_path
            if fulloc:
                path = self.full_path
            if self.yamlexists:
                self.yamlremove()
                print(f"Removed {path!r}")
            else:
                raise FileNotFoundError(
                    f"the state file {path!r} is missing"
                )
        if key and not value:
            key_value = self.get_value(key)
            state: dict[Key, Any] = {key: key_value}
            if key_value is not None:
                print(f"Removed {state} from {self.file_name!r}")
                self.yamlchange(key, None)
            else:
                raise KeyValueError(
                    f"{key!r} was not nullified {state}, no value exists for defined key"
                )
        if key and value:
            key_value = self.get_value(key)
            state: dict[Key, Any] = {key: value}
            if key_value == value:
                print(f"Removed {state} from {self.file_name!r}")
                self.yamlchange(key, None)
            else:
                raise KeyValueError(
                    f"{key!r} was not nullified {state}, value ({value!r}) does not match"
                )

    def manager(self, key: Key, value: Value):
        state: dict[Key, Any] = {key: value}
        print(f"State Changed {state}")
        if self.yamlexists and self.get_value(key):
            self.yamlmodel(state, write=True, append=True)
        else:
            self.yamlmodel(state, write=True)

    def _modifier(
        self,
        value: Value,
        operation: Literal[0] | Literal[1] | Literal[2] | Literal[3],
        *,
        batch: int,
    ):
        check = value
        if value is not None:
            check = (
                value + self.multiplier
                if operation == 0
                else value - self.multiplier
            )
            if operation == 2:
                check = value + batch
            if operation == 3:
                check = value - batch
            if check % batch != 0:
                check = f"{check}, batch % {batch} != 0"
        if check is None:
            check = f"±0, {{{operation}, {value}*{self.multiplier}}}"
        if isinstance(self.limit, tuple) or isinstance(
            self.ilimit, tuple
        ):
            inclusive_limit = range(self.ilimit[0], self.ilimit[1] + 1)
            range_limit = (
                range(*self.limit)
                if not self.ilimit
                else inclusive_limit
            )
            if value not in range_limit or check not in range_limit:
                if self.limit:
                    raise RangeError(
                        f"Index ({check}) not in", range(*self.limit)
                    )
                if self.ilimit:
                    raise RangeError(
                        f"Index ({check}) not in", range(*self.ilimit)
                    )
                cancel()
        if isinstance(self.limit, int) or isinstance(self.ilimit, int):
            range_limit = (
                range(self.limit)
                if not self.ilimit
                else range(self.ilimit + 1)
            )
            if value not in range_limit or check not in range_limit:
                if self.limit:
                    raise RangeError(
                        f"Index ({check}) not in", range(self.limit)
                    )
                if self.ilimit:
                    raise RangeError(
                        f"Index ({check}) not in", range(self.ilimit)
                    )
                cancel()
        return True

    def decr(self, key: Key):
        key_value = self.get_value(key)
        old_state: dict[Key, Any] = {key: key_value}
        state: dict[Key, Any] = {}
        if isinstance(key_value, int):
            state: dict[Key, Any] = {key: key_value - self.multiplier}
        try:
            smdecr = self._modifier(key_value, 1, batch=self.multiplier)
            if state[key] != key_value and smdecr:
                print(f"Decremented {old_state} -> {state}")
                self.yamlmodel(state, write=True, append=True)
        except (KeyValueError, RangeError) as e:
            print(e)

    def incr(self, key: Key):
        key_value = self.get_value(key)
        old_state: dict[Key, Any] = {key: key_value}
        state: dict[Key, Any] = {}
        if isinstance(key_value, int):
            state: dict[Key, Any] = {key: key_value + self.multiplier}
        try:
            smincr = self._modifier(key_value, 0, batch=self.multiplier)
            if state[key] != key_value and smincr:
                print(f"Incremented {old_state} -> {state}")
                self.yamlmodel(state, write=True, append=True)
        except (KeyValueError, RangeError) as e:
            print(e)

    def bypass(self, key: Key, batch: int):
        key_value = self.get_value(key)
        old_state: dict[Key, Any] = {key: key_value}
        state: dict[Key, Any] = {}
        if isinstance(key_value, int):
            state: dict[Key, Any] = {key: key_value + batch}
        try:
            smbatch = self._modifier(key_value, 2, batch=batch)
            if state[key] != key_value and smbatch:
                print(
                    f"Incremented {old_state} -> (batch={batch}) -> {state}"
                )
                self.yamlmodel(state, write=True, append=True)
        except (KeyValueError, RangeError) as e:
            print(e)

    def nbypass(self, key: Key, batch: int):
        key_value = self.get_value(key)
        old_state: dict[Key, Any] = {key: key_value}
        state: dict[Key, Any] = {}
        if isinstance(key_value, int):
            state: dict[Key, Any] = {key: key_value - batch}
        try:
            smbatch = self._modifier(key_value, 3, batch=batch)
            if state[key] != key_value and smbatch:
                print(
                    f"Decremented {old_state} -> (batch={batch}) -> {state}"
                )
                self.yamlmodel(state, write=True, append=True)
        except (KeyValueError, RangeError) as e:
            print(e)

    def toggle(self, key: Key, *, on: bool = False, off: bool = False):
        self.multiplier = 2
        self.ilimit = (-1, 1)
        key_value = self.get_value(key)
        old_state: dict[Key, Any] = {
            key: "[green]on[/]" if key_value == 1 else "[red]off[/]"
        }
        state: dict[Key, Any] = {}
        state_repr: dict[Key, Any] = {}
        if on:
            if isinstance(key_value, int):
                state: dict[Key, Any] = {key: 1}
                state_repr: dict[Key, Any] = {key: "[green]on[/]"}
            try:
                if state[key] != key_value:
                    print(f"Toggled {old_state} -> {state_repr}")
                    self.yamlmodel(state, write=True, append=True)
            except (KeyValueError, RangeError) as e:
                print(e)
        if off:
            if isinstance(key_value, int):
                state: dict[Key, Any] = {key: -1}
                state_repr: dict[Key, Any] = {key: "[red]off[/]"}
            try:
                if state[key] != key_value:
                    print(f"Toggled {old_state} -> {state_repr}")
                    self.yamlmodel(state, write=True, append=True)
            except (KeyValueError, RangeError) as e:
                print(e)



# ----------------------------------------------------------------------
# |  Simple-Directories
# ----------------------------------------------------------------------
from montyalex.console_tools import richconsole, success_stm
from montyalex.fs_tools import (
    current_working_dir as cwd,
    joinpaths,
    mkdirs,
    pathexists,
)
from .helpdirs import remove_formatted__dir

print = richconsole.print


def create_simple_directories(
    range_: int = 1,
    name_: str = None,
    prefix: str = None,
    suffix: str = None,
    silent: bool = False,
):
    for i in range(range_):
        formatted = name_ if name_ else f"{i:03}"
        if prefix:
            formatted = f"{prefix}{formatted}"
        if suffix:
            formatted = f"{formatted}{suffix}"
        mkdirs(joinpaths(cwd, formatted), exist_ok=True)
        if not silent:
            print(f"{success_stm}, Created {formatted!r} in {cwd!r}")


def remove_simple_directories(
    range_: int = 1,
    name_: str = None,
    prefix: str = None,
    suffix: str = None,
    silent: bool = False,
):
    for i in range(range_):
        formatted = name_ if name_ else f"{i:03}"
        if prefix:
            formatted = f"{prefix}{formatted}"
        if suffix:
            formatted = f"{formatted}{suffix}"
        formatted_directory = joinpaths(cwd, formatted)
        remove_formatted__dir(formatted_directory, silent)

        if not pathexists(formatted_directory) and not silent:
            print(f"{success_stm}, Removed {formatted!r} from {cwd!r}")



from montyalex.fs_tools import cancel


def date_directories(
    name_: str, prefix: str, suffix: str, silent: bool = False
):
    if (name_ or prefix or suffix) and not silent:
        print("Not allowed with the -datedirs option")
        cancel()



from montyalex.uo_tools import schema, json, toml, yaml, mpck
from .init_objects import settings_schema_definition
from .uo_init import initialize_useroptions


def initialize_schema(dirname: str = ".mtax", filename: str = "schema"):
    _schema = schema(directory=dirname, filename=filename)
    _schema.change("$schema", "http://json-schema.org/draft-07/schema#")
    _schema.change("$id", "montyalex.python-cli.settings-v0.1.4")
    _schema.change("title", "Settings")
    _schema.change("description", "MontyAlex CLI Settings")
    _schema.change("type", "object")
    _schema.change("properties", {"mtax": settings_schema_definition})
    _schema.change("required", ["mtax"])


def initialize_void(
    uo: json | toml | yaml | mpck,
    dirname: str,
    filename: str,
    *,
    use_all: bool,
    use_synced: bool,
    use_json: bool,
    use_toml: bool,
    use_yaml: bool,
    use_mpck: bool,
):
    if not uo.exists:
        print("Creating new settings file...")
        initialize_useroptions(
            dirname=dirname,
            filename=filename,
            overwrite=False,
            use_all=use_all,
            use_synced=use_synced,
            use_json=use_json,
            use_toml=use_toml,
            use_yaml=use_yaml,
            use_mpck=use_mpck,
        )



# ----------------------------------------------------------------------
# |  MontyAlex
# ----------------------------------------------------------------------
from .cache_tools import (
    MtaxCache,
    get_cache_item,
    add_cache_item,
    set_cache_item,
    incr_cache_item,
    decr_cache_item,
    remove_cache_item,
    cache_info,
    cache_list,
    cache_clear,
    cache_reset,
)
from .console_tools import (
    richconsole,
    success_stm,
    info_stm,
    warn_stm,
    error_stm,
    debug_stm,
    critical_stm,
)
from .directory_tools import (
    datedirs,
    rmdatedirs,
    simpledirs,
    rmsimpledirs,
)
from .future_tools import __november24__, __december24__
from .time_tools import MtaxTime


__all__ = [
    "MtaxCache",
    "get_cache_item",
    "add_cache_item",
    "set_cache_item",
    "incr_cache_item",
    "decr_cache_item",
    "remove_cache_item",
    "cache_info",
    "cache_list",
    "cache_clear",
    "cache_reset",
    "richconsole",
    "success_stm",
    "info_stm",
    "warn_stm",
    "error_stm",
    "debug_stm",
    "critical_stm",
    "datedirs",
    "rmdatedirs",
    "simpledirs",
    "rmsimpledirs",
    "__november24__",
    "__december24__",
    "MtaxTime",
]



# ----------------------------------------------------------------------
# |  Time-Cache App
# ----------------------------------------------------------------------
from montyalex.time_tools import MtaxTime
from montyalex.typer_tools import Option, Typer
from .mtax_cache import MtaxCache
from .zonecache_app import zone_


# ----------------------------------------------------------------------
# |  Typer App
# ----------------------------------------------------------------------
time_: Typer = Typer(name="time", help="Manage time-data in the cache")
time_.add_typer(zone_)


# ----------------------------------------------------------------------
# |  Typer Commands; reset_, set_, get_
# ----------------------------------------------------------------------
@time_.command(name="reset")
def reset_(
    *,
    timezone: str = "Etc/Greenwich",
    silent: bool = Option(
        False,
        "-s",
        show_default="-!s",
        help="Silence extraneous messages in console",
    ),
):
    """Reset timestamp in the cache"""
    mtax: MtaxCache = MtaxCache(timezone)
    mtax.remove("_timestamp_", silent)
    set_(timezone=timezone, silent=silent)


@time_.command(name="set")
def set_(
    *,
    timezone: str = "Etc/Greenwich",
    silent: bool = Option(
        False,
        "-s",
        show_default="-!s",
        help="Silence extraneous messages in console",
    ),
):
    """Set timestamp in the cache"""
    mtax: MtaxCache = MtaxCache(timezone)
    mtax.set_item(
        ("_timestamp_", MtaxTime(timezone).timestamp()), silent
    )


@time_.command(name="get")
def get_(
    *,
    timezone: str = "Etc/Greenwich",
    silent: bool = Option(
        False,
        "-s",
        show_default="-!s",
        help="Silence extraneous messages in console",
    ),
):
    """Set timestamp in the cache"""
    mtax: MtaxCache = MtaxCache(timezone)
    mtax.get_item("_timestamp_", silent)



# ----------------------------------------------------------------------
# |  Virtual Memory Process
# ----------------------------------------------------------------------
from .virtual_memory import VirtualMemory


class VirMemProcess:
    """
    Management for virtual memory processes with context capabilities

    Attrs:
        available (bool): Flag to set recommended_maximum to available memory.
        double (bool): Flag to double the recommended_maximum.
        maximum (bool): Flag to set recommended_maximum to total system memory.
        usage (bool): Flag to print memory usage details.
        virmem (VirtualMemory): An instance of the VirtualMemory class.
        system (dict): The system memory details from the VirtualMemory instance.

    Methods:
        __call__: Updates the flags for managing memory.
        __enter__: Context manager entry, configures memory settings based on flags.
        __exit__: Context manager exit, cleans up resources.
    """

    def __init__(self) -> None:
        self.available: bool = False
        self.double: bool = False
        self.maximum: bool = False
        self.usage: bool = False
        self.virmem = VirtualMemory()
        self.system = self.virmem.system

    def __call__(
        self,
        *,
        available: bool = False,
        double: bool = False,
        maximum: bool = False,
        usage: bool = False,
    ) -> None:
        self.available: bool = available
        self.double: bool = double
        self.maximum: bool = maximum
        self.usage: bool = usage
        self.virmem = VirtualMemory()
        return self

    def __enter__(self):
        self.virmem()

        if self.available:
            self.system["recommended_maximum"] = self.system[
                "available"
            ]
        if self.double:
            self.system["recommended_maximum"] *= 2
        if self.maximum:
            self.system["recommended_maximum"] = self.system["total"]
        if self.usage:
            print(f'System Used: {self.system["used"]:.2f}gb')
            print(f'System Available: {self.system["available"]:.2f}gb')
            print(f'System Total: {self.system["total"]:.2f}gb')
            print(
                f'Recommended Maximum: {self.system["recommended_maximum"]:.2f}gb'
            )
            print(f"Process Used: {self.virmem.rss / 2:.2f}gb")
        if float(self.virmem) > self.virmem.rss / 2:
            return self
        raise MemoryError(
            "Context manager not allowed due to insufficient available memory"
        )

    def __exit__(self, type, value, traceback):
        pass



# ----------------------------------------------------------------------
# |  Cache Tools
# ----------------------------------------------------------------------
from .mtax_cache import MtaxCache
from .cache_app import (
    cache_,
    get_ as get_cache_item,
    add_ as add_cache_item,
    set_ as set_cache_item,
    incr_ as incr_cache_item,
    decr_ as decr_cache_item,
    remove_ as remove_cache_item,
    info_ as cache_info,
    list_ as cache_list,
    clear_ as cache_clear,
    reset_ as cache_reset)


__version__ = 'v1.0.0'


__all__ = [
    "MtaxCache",
    "cache_",
    "get_cache_item",
    "add_cache_item",
    "set_cache_item",
    "incr_cache_item",
    "decr_cache_item",
    "remove_cache_item",
    "cache_info",
    "cache_list",
    "cache_clear",
    "cache_reset",
    "__version__"]



# ----------------------------------------------------------------------
# |  Timing
# ----------------------------------------------------------------------
import concurrent.futures
import time

from montyalex.fs_tools import cancel
from montyalex.us_tools import SETTINGS


TIMEOUT = "1min"
if SETTINGS.get_value("default.opt.timeout"):
    TIMEOUT = SETTINGS.get_value("default.opt.timeout")
MINUTE = False
SECONDS = False

if isinstance(TIMEOUT, str):
    if "min" in TIMEOUT:
        MINUTE = True
        TIMEOUT = int(TIMEOUT.removesuffix("min")) * 60
if isinstance(TIMEOUT, str):
    if "sec" in TIMEOUT:
        SECONDS = True
        TIMEOUT = int(TIMEOUT.removesuffix("sec"))


# ----------------------------------------------------------------------
# |  Function Time Decorator with Timeout
# ----------------------------------------------------------------------
def func_time(message: str = None):
    def decorator(func):
        func_name = f'"{func.__name__}({func.__type_params__})"'

        def wrapper(*args, **kwargs):
            local_message = (
                message
                if message is not None
                else "completed in approximately"
            )

            start_time = time.time()
            with concurrent.futures.ThreadPoolExecutor() as executor:
                future = executor.submit(func, *args, **kwargs)
                try:
                    result = future.result(timeout=TIMEOUT)
                    elapsed_time = time.time() - start_time

                    if local_message and "+" in local_message:
                        func_msg_1, func_msg_2 = [
                            part.strip()
                            for part in local_message.split(
                                "+", maxsplit=1
                            )
                        ]
                        print(
                            f"{func_name}: {func_msg_1} {elapsed_time:.3f} seconds {func_msg_2}"
                        )
                    elif local_message and ";" in local_message:
                        func_msg_1, func_msg_2 = [
                            part.strip()
                            for part in local_message.split(
                                ";", maxsplit=1
                            )
                        ]
                        print(
                            f"{func_name}: {func_msg_1} {elapsed_time:.3f} seconds {func_msg_2}"
                        )
                    else:
                        print(
                            f"{func_name}: {local_message} {elapsed_time:.3f} seconds"
                        )

                    return result

                except concurrent.futures.TimeoutError:
                    print(
                        f"Function {func_name} exceeded {TIMEOUT} seconds. Cancelling..."
                    )
                    cancel("Cancelled due to timeout")

        return wrapper

    return decorator


# ----------------------------------------------------------------------
# |  Example Usage
# ----------------------------------------------------------------------
if __name__ == "__main__":

    @func_time(message="completed in")
    def short_running_action():
        time.sleep(2)
        print("SHORT Action completed")

    @func_time(message="took + to complete")
    def long_running_action():
        time.sleep(4)
        print("LONG Action completed")

    @func_time(message="took; to complete!")
    def longer_running_action():
        time.sleep(8)
        print("LONGER Action completed")

    @func_time()
    def longest_running_action():
        time.sleep(16)
        print("LONGEST Action completed")

    short_running_action()
    long_running_action()
    longer_running_action()
    longest_running_action()



# ----------------------------------------------------------------------
# |  Virtual Memory
# ----------------------------------------------------------------------
import os
import psutil
from .memory_utils import virtualmem


class VirtualMemory:
    """
    VirtualMemory management

    Attrs:
        system (dict): A dict containing total, available, used, and recommended maximum memory.
        pid (int): The process ID of the current Python process.
        process (psutil.Process): The current Python process.
        memory_info (psutil.Process.memory_info): The memory info of the current Python process.
        used (float): The used memory of the current Python process in MB.
        rss (float): The resident set size of the current Python process in MB.

    Methods:
        __call__: Updates the system memory and RSS values.
        __float__: Returns the recommended maximum memory in MB.
    """

    def __init__(self) -> None:
        self.system = virtualmem()
        self.pid = os.getpid()
        self.process = psutil.Process(self.pid)
        self.memory_info = self.process.memory_info()
        self.used = self.system["used"]
        self.rss = self.memory_info.rss / (1024**2) - self.used

    def __float__(self) -> float:
        return self.system["recommended_maximum"]

    def __call__(self) -> "VirtualMemory":
        self.system = virtualmem()
        self.memory_info = self.process.memory_info()
        self.rss = self.memory_info.rss / (1024**2) - self.used
        return self



# ----------------------------------------------------------------------
# |  _User-Info
# ----------------------------------------------------------------------
from ._components import (
    GroupSettingName,
    ShortSettingName,
    SettingName,
    SettingComponent,
)


# ----------------------------------------------------------------------
# |  User Info Setting Component
# ----------------------------------------------------------------------
class UserInfo(SettingComponent):
    """A UserInfo setting component represents a specific user-related setting.

    It is defined by an infix, suffix, type, and optionally a prefix and parent.

    The prefix and infix are used to construct the setting's identifier.
    The parent is an instance of a setting name class that provides hierarchical context.
    The type is used to validate associated values.
    """

    def __init__(
        self,
        infix: str,
        suffix: str,
        type_: type,
        *,
        prefix: str = "user",
        parent: (
            GroupSettingName | ShortSettingName | SettingName
        ) = None,
    ) -> None:
        super().__init__(
            prefix, infix, suffix, type_=type_, parent=parent
        )


# ----------------------------------------------------------------------
# |  User Locale Instances
# ----------------------------------------------------------------------
locale_continent_userinfo = UserInfo(
    "locale",
    "continent",
    str,
    parent=SettingName("user", "locale", "continent"),
)
locale_country_userinfo = UserInfo(
    "locale",
    "country",
    str,
    parent=SettingName("user", "locale", "country"),
)
locale_region_userinfo = UserInfo(
    "locale",
    "region",
    str,
    parent=SettingName("user", "locale", "region"),
)
locale_state_userinfo = UserInfo(
    "locale",
    "state",
    str,
    parent=SettingName("user", "locale", "state"),
)
locale_province_userinfo = UserInfo(
    "locale",
    "province",
    str,
    parent=SettingName("user", "locale", "province"),
)
locale_city_userinfo = UserInfo(
    "locale", "city", str, parent=SettingName("user", "locale", "city")
)
locale_streetname_userinfo = UserInfo(
    "locale",
    "street-name",
    str,
    parent=SettingName("user", "locale", "street-name"),
)
locale_streetnum_userinfo = UserInfo(
    "locale",
    "street-number",
    str,
    parent=SettingName("user", "locale", "street-number"),
)



# ----------------------------------------------------------------------
# |  File-System Tools
# ----------------------------------------------------------------------
from .path import readfile
from .paths import (
    abspath,
    expanduser,
    joinpaths,
    mkdirs,
    rmdirs,
    rmfile,
    rmtree,
    pathexists,
    current_working_dir,
)
from .system import cancel


__version__ = "v1.0.0"


__all__ = [
    "readfile",
    "abspath",
    "expanduser",
    "joinpaths",
    "mkdirs",
    "rmdirs",
    "rmfile",
    "rmtree",
    "pathexists",
    "current_working_dir",
    "cancel",
    "__version__",
]



# ----------------------------------------------------------------------
# |  Path
# ----------------------------------------------------------------------
from pathlib import Path


def current_path(path=".") -> Path:
    """Returns the current path"""
    return Path(path)


def current_working_path(path: Path) -> Path:
    """Returns the current working directory path"""
    return current_path(path).resolve()


# ----------------------------------------------------------------------
# |  Appending paths
# ----------------------------------------------------------------------
def appendp(
    path: str, *others: str, current_directory: bool = False
) -> Path:
    """Append paths to another path"""
    _path = current_path(path).joinpath(*map(str, others))
    return current_working_path(_path) if current_directory else _path


def appendf(
    path: str,
    *others: str,
    filename: str,
    suffix: str = None,
    current_directory: bool = False,
) -> Path:
    """Append a filename to a path, with an optional suffix."""
    return appendp(
        path,
        *others,
        f"{filename}.{suffix}" if suffix else filename,
        current_directory=current_directory,
    )


def isabsolute(path: str) -> bool:
    """Returns True if the path is absolute, False otherwise."""
    return Path(path).is_absolute()


def isrelative(path: str) -> bool:
    """Returns True if the path is relative, False otherwise."""
    return not isabsolute(path)


def isfile(path: str) -> bool:
    """
    Returns True if the path refers to a regular file (not a directory).
    """
    return Path(path).is_file()


def readfile(path: str) -> str:
    """
    Reads the contents of a file at the given path, decoding with UTF-8.
    """
    return Path(path).read_text(encoding="UTF-8")


def pathexists(path: str, debug_cwd: bool = False) -> bool:
    """
    Returns True if the path refers to an existing path.
    If debug_cwd is True, prints the absolute path.
    """
    if debug_cwd:
        path = current_working_path(path)
        print(path)
    abs = isabsolute(path)
    return Path(path).exists() if abs else False


# ----------------------------------------------------------------------
# |  Example Usage
# ----------------------------------------------------------------------
if __name__ == "__main__":
    for _ in range(100):
        appended = appendf(
            "src",
            "montyalex",
            "fs_tools",
            f"{_:02}",
            filename="paths",
            suffix="py",
            current_directory=True,
        )
        E = pathexists(appended)
        absolute = isabsolute(appended)
        relative = isrelative(appended)
        print(appended, E, absolute, relative)



# ----------------------------------------------------------------------
# |  Singleton
# ----------------------------------------------------------------------
def singleton_decorator(get_identifier=None):
    def wrapper(cls):
        instances = {}

        def get_instance(*args, **kwargs):
            if get_identifier:
                identifier = get_identifier(*args, **kwargs)
            else:
                identifier = None
            key = (cls, identifier)
            if key not in instances:
                instances[key] = cls(*args, **kwargs)
            return instances[key]

        return get_instance

    return wrapper



# ----------------------------------------------------------------------
# |  Future Tools
# ----------------------------------------------------------------------
import importlib
from inspect import isclass, isfunction

from montyalex.typing_tools import Callable, Type, Union, Dict


T = Dict[str, Union[Type[object], Callable]]


class MonthlyFeatures:
    _monthly_feature_map = {
        "2024_11": "montyalex.future_tools.by_2024_11_27",
        "2024_12": "montyalex.future_tools.by_2024_12_25",
    }

    def __init__(self, release_date: str):
        self.release_date = release_date
        self.features: T = self.load_month(release_date)

    def load_month(self, release_date: str) -> T:
        module_name = self._monthly_feature_map.get(release_date[:7])
        if not module_name:
            return {}

        module = importlib.import_module(module_name)

        features = {
            name: getattr(module, name)
            for name in dir(module)
            if (
                isclass(getattr(module, name))
                or isfunction(getattr(module, name))
            )
            and getattr(module, name).__module__ == module_name
        }

        return features

    def __getattr__(self, item: str):
        if item in self.features:
            return self.features[item]
        raise AttributeError(
            f"'{self.__class__.__name__}' object has no attribute '{item}'"
        )

    @property
    def exceptions(self) -> Dict[str, Type[BaseException]]:
        return {
            name: feature
            for name, feature in self.features.items()
            if isclass(feature) and issubclass(feature, BaseException)
        }

    @property
    def functions(self) -> Dict[str, Callable]:
        return {
            name: feature
            for name, feature in self.features.items()
            if isfunction(feature)
        }

    @property
    def other_classes(self) -> Dict[str, Type[object]]:
        return {
            name: feature
            for name, feature in self.features.items()
            if isclass(feature)
            and not issubclass(feature, BaseException)
        }

    @property
    def matching_features(self, keyword: str) -> T:
        return {
            name: feature
            for name, feature in self.features.items()
            if keyword.lower() in name.lower()
        }


# ----------------------------------------------------------------------
# |  Global Future Module Objects
# ----------------------------------------------------------------------
__november24__ = MonthlyFeatures("2024_11_27")
__december24__ = MonthlyFeatures("2024_12_25")

__version__ = "v1.0.0"


__all__ = ["__november24__", "__december24__", "__version__"]



# ----------------------------------------------------------------------
# |  Object Tools
# ----------------------------------------------------------------------
from .enumeration import Key, Value
from .exceptions import (
    CacheError,
    ConsoleError,
    DirectoryError,
    FileSystemError,
    KeyValueError,
    JSONError,
    RangeError,
    SettingsError,
    TimeError,
    TyperAppError,
    UserOptionsError,
    UserSettingsError,
)
from .pickling import pdumps, ploads
from .singleton import singleton_decorator as singleton


__version__ = "v1.0.0"


__all__ = [
    "Key",
    "Value",
    "CacheError",
    "ConsoleError",
    "DirectoryError",
    "FileSystemError",
    "KeyValueError",
    "JSONError",
    "RangeError",
    "SettingsError",
    "TimeError",
    "TyperAppError",
    "UserOptionsError",
    "UserSettingsError",
    "pdumps",
    "ploads",
    "singleton",
    "__version__",
]



# ----------------------------------------------------------------------
# |  MontyAlex Cache
# ----------------------------------------------------------------------
import platform
from diskcache import Cache
import psutil

from montyalex.console_tools import richconsole, success_stm
from montyalex.fs_tools import (
    current_working_dir,
    joinpaths,
    pathexists,
)
from montyalex.time_tools import MtaxTime
from montyalex.typing_tools import NoneType
from .cache_helpers import remove_cache__dir

print = richconsole.print


DIR = joinpaths(".mtax", "cache")
mtax: Cache = Cache(DIR)
os_name = platform.system()
os_strength = platform.processor()
py_ver = platform.python_version()
cache_exists = joinpaths(current_working_dir, DIR, "cache.db")

svmem = psutil.virtual_memory()

total_ram_bytes = svmem.total
total_ram_gb = total_ram_bytes / (1024**3)

available_ram_bytes = svmem.available
available_ram_gb = available_ram_bytes / (1024**3)

used_ram_bytes = total_ram_bytes - available_ram_bytes
used_ram_gb = used_ram_bytes / (1024**3)


def style(
    s: str,
    *,
    color: str = "default",
    bold: bool = False,
    indent: int = 4,
) -> str:
    styled = f"[{color}]{s}[/]"
    if bold:
        styled = f"[bold]{styled}[/]"
    indented_styled = f'{" " * indent}{styled}'
    return indented_styled


class MtaxCache:
    """
    A class to manage a disk cache.

    The cache is stored in a SQLite database in the current working
    directory. The cache is keyed by strings, and the values are any
    Python object that can be pickled.

    The cache also stores some information about the current Python
    interpreter, such as the version, the operating system, and the
    amount of available RAM.

    The cache can be cleared, and individual items can be removed.

    The cache can be listed, showing all the key-value pairs.

    The cache can be reset, which clears the cache and sets the
    Python interpreter information to the current values.

    The cache can be checked, which shows information about the
    current Python interpreter, and checks if the cache directory
    exists.
    """

    def __init__(self, timezone: str = "Etc/Greenwich") -> None:
        self.cache = mtax
        self.status: str = "Ok"
        self.os_ = os_name
        self.os_strength = os_strength
        self.py_ver = py_ver
        self.total_ram_gb = total_ram_gb
        self.available_ram_gb = available_ram_gb
        self.used_ram_gb = used_ram_gb
        self.datetime: MtaxTime = MtaxTime(timezone)

    def add_item(self, item: tuple[str, str], silent: bool = False):
        if not silent:
            print(f"Adding {item[0]!r} to Cache")
        self.cache.add("_timestamp_", self.datetime.timestamp())
        self.cache.add(item[0], item[1])
        if not silent:
            print(f"{success_stm}, Added {item[0]!r} to Cache")

    def set_item(self, item: tuple[str, str], silent: bool = False):
        to_message = f"{item[0]!r} to {item[1]!r}"
        if not silent:
            print(f"Setting {to_message} in the Cache")
        self.cache.set("_timestamp_", self.datetime.timestamp())
        self.cache.set(item[0], item[1])
        print(f"{success_stm}, Set {to_message} in the Cache")

    def get_item(self, key: str, silent: bool = False):
        if not silent:
            print(f"Getting {key!r} from the Cache")
        item_value = self.cache.get(key)
        if not silent:
            print(f"{success_stm}, Got {key!r} from the Cache")
            print(
                f"The value for {key!r} in the Cache is {item_value!r}"
            )
        if not silent:
            print(f"{key!r}={item_value!r}")
        return item_value

    def incr_item(self, key: str, silent: bool = False):
        if not silent:
            print(f"Incrementing {key!r} in the Cache")
        value = self.cache.incr(key)
        value = max(value, 0)
        if not silent:
            print(
                f"{success_stm}, Incremented {key!r}"
                f" in the Cache to {value}"
            )
        return value

    def decr_item(self, key: str, silent: bool = False):
        if not silent:
            print(f"Decrementing {key!r} in the Cache")
        value = self.cache.decr(key)
        value = max(value, 0)
        if not silent:
            print(
                f"{success_stm}, Decremented {key!r}"
                f" in the Cache to {value}"
            )
        return value

    def clear(self, silent: bool = False):
        if not silent:
            print("Clearing Cache")
        self.cache.clear()
        remove_cache__dir(DIR, silent)
        if not silent:
            print("Cache Cleared")

    def reset(self, silent: bool = False):
        if not silent:
            print("Resetting Cache")
        self.remove("_python_", silent)
        self.remove("_datedirs_", silent)
        self.remove("_cache-tools_", silent)
        self.remove("_directory-tools_", silent)
        self.add_item(
            ("_python_", f"v{platform.python_version()}"), silent
        )
        self.add_item(("_datedirs_", "v1.0.0"), silent)
        self.add_item(("_cache-tools_", "v1.0.0"), silent)
        self.add_item(("_directory-tools_", "v1.0.0"), silent)
        if not silent:
            print("Cache Reset")

    def remove(self, key: str, silent: bool = False):
        if not silent:
            print(f"Removing {key!r}")
        self.cache.delete(key)
        if not silent:
            print(f"{success_stm}, Removed {key!r} from Cache")

    def list_k_v_pairs(self, silent: bool = False):
        cached_items = list(self.cache)
        cached_items.sort()
        for key in cached_items:
            value = self.cache.get(key)
            print(f"[bold]{key}[/]: {value!r}")
        if not silent:
            ...

    def info(self, list_: bool = False, silent: bool = False):
        current_time = f"{self.datetime.timestamp()!r}"
        current_status = f"{self.status!r}"
        current_strength = f"{self.os_strength!r}"
        current_os = f"{self.os_!r}"

        current_py_ver = f"[yellow]'v{self.py_ver}'[/]"

        current_aram = f"[sky_blue2]{self.available_ram_gb:.2f}GB[/]"
        none_aram = f"[orange4]{0:.2f}GB [bold]???[/][/] NONE FOUND"

        current_uram = f"[deep_pink3]{self.used_ram_gb:.2f}GB[/]"
        none_uram = f"[green1]{0:.2f}GB [bold]???[/][/] NONE FOUND"

        current_tram = f"[sky_blue2]{self.total_ram_gb:.2f}GB[/]"
        none_tram = f"[orange4]{0:.2f}GB [bold]???[/][/] NONE FOUND"
        print(f"{style('Info', bold=True)}: [")
        print(
            style(
                f"{style(
                    'Current',
                    bold=True)}: {current_time}",
                indent=2,
            )
        )
        if self.status == "Ok":
            print(
                style(
                    f"{style(
                        'Status',
                        bold=True)}: [green]{current_status}[/]",
                    indent=2,
                )
            )
        else:
            print(
                style(
                    f"{style(
                        'Status',
                        bold=True)}: [red]{current_status}[/]",
                    indent=2,
                )
            )
        print(
            style(
                f"{style(
                    'System Processor',
                    bold=True)}: {current_strength}",
                indent=2,
            )
        )
        print(
            style(
                f"{style(
                    'Operating System',
                    bold=True)}: {current_os}",
                indent=2,
            )
        )
        if self.os_strength and self.os_:
            print(
                style(
                    f"{style(
                        'Python Version',
                        bold=True)}: {current_py_ver}",
                    indent=4,
                )
            )
            if self.available_ram_gb:
                print(
                    style(
                        f"{style(
                            'RAM Available',
                            bold=True)}: {current_aram}",
                        indent=4,
                    )
                )
            else:
                print(
                    style(
                        f"{style(
                            'RAM Available',
                            bold=True)}: {none_aram}",
                        indent=4,
                    )
                )
            if self.used_ram_gb:
                print(
                    style(
                        f"{style(
                            'RAM Used',
                            bold=True)}: {current_uram}",
                        indent=4,
                    )
                )
            else:
                print(
                    style(
                        f"{style(
                            'RAM Used',
                            bold=True)}: {none_uram}",
                        indent=4,
                    )
                )
            if self.total_ram_gb > 0:
                print(style(f"RAM Total: {current_tram}", indent=4))
            else:
                print(style(f"RAM Total: {none_tram}", indent=4))
            print(style("Status: [green]'Good'[/]", indent=4))
        else:
            print(style("Status: [red]'Bad'[/]", indent=4))
        if pathexists(cache_exists):
            print(
                style(
                    f"{style(
                        'Cache Directory',
                        bold=True)}: [green]{DIR!r}[/]",
                    indent=2,
                )
            )
            print(
                style(
                    f"{style(
                        'File',
                        bold=True)}: [green]'cache.db'[/]",
                    indent=4,
                )
            )
        else:
            print(
                style(
                    f"{style(
                    'Cache Directory',
                    bold=True)}: [red]{DIR!r}[/]",
                    indent=2,
                )
            )
            print(
                style(
                    f"{style(
                        'File',
                        bold=True)}: [red]'cache.db'[/] [dim](does not exist)[/]",
                    indent=4,
                )
            )
        py_ver_cached = self.get_item("_python_", True)
        cat_ver_cached = self.get_item("_cache-tools_", True)
        cot_ver_cached = self.get_item("_console-tools_", True)
        dir_ver_cached = self.get_item("_directory-tools_", True)
        if isinstance(py_ver_cached, NoneType):
            print(
                style(
                    f"{style(
                    'Python Version',
                    bold=True)}: {py_ver_cached!r}",
                    indent=4,
                )
            )
        else:
            print(
                style(
                    f"{style(
                    'Python Version',
                    bold=True)}: [yellow]{py_ver_cached!r}[/]",
                    indent=4,
                )
            )
        if isinstance(cat_ver_cached, NoneType):
            print(
                style(
                    f"{style(
                    'CacheTools Version',
                    bold=True)}: {cat_ver_cached!r}",
                    indent=4,
                )
            )
        else:
            print(
                style(
                    f"{style(
                    'CacheTools Version',
                    bold=True)}: [cyan2]{cat_ver_cached!r}[/]",
                    indent=4,
                )
            )
        if isinstance(cot_ver_cached, NoneType):
            print(
                style(
                    f"{style(
                    'ConsoleTools Version',
                    bold=True)}: {cot_ver_cached!r}",
                    indent=4,
                )
            )
        else:
            print(
                style(
                    f"{style(
                    'ConsoleTools Version',
                    bold=True)}: [cyan2]{cot_ver_cached!r}[/]",
                    indent=4,
                )
            )
        if isinstance(dir_ver_cached, NoneType):
            print(
                style(
                    f"{style(
                    'DirectoryTools Version',
                    bold=True)}: {dir_ver_cached!r}",
                    indent=4,
                )
            )
        else:
            print(
                style(
                    f"{style(
                    'DirectoryTools Version',
                    bold=True)}: [cyan2]{dir_ver_cached!r}[/]",
                    indent=4,
                )
            )
        if list_:
            print("]..,")
            self.list_k_v_pairs()
        else:
            print("]")
        if not silent:
            ...



# ----------------------------------------------------------------------
# |  MontyAlex Time
# ----------------------------------------------------------------------
from datetime import datetime
from zoneinfo import ZoneInfo, ZoneInfoNotFoundError


# ----------------------------------------------------------------------
# |  Timezone Based Datetime Formatting
# ----------------------------------------------------------------------
class MtaxTime:
    def __init__(self, timezone: str = "Etc/Greenwich") -> None:
        self.timezone: str = timezone
        try:
            self.tzinfo: ZoneInfo = ZoneInfo(self.timezone)
        except ZoneInfoNotFoundError as exc:
            raise ValueError(
                f"Invalid timezone name: {self.timezone}"
            ) from exc
        self.now: datetime = datetime.now(self.tzinfo)

        self.year: int = self.now.year
        self.month: int = self.now.month
        self.day: int = self.now.day
        self.name: str = self.now.strftime("%a").upper()
        self.daynum_ofweek: int = self.now.weekday()
        self.daynum_ofyear: int = self.now.timetuple().tm_yday

        self.hour: str = f"{self.now.hour:02}"
        self.minute: str = f"{self.now.minute:02}"
        self.second: str = f"{self.now.second:02}"
        self.microsecond: int = self.now.microsecond
        self.millisecond: int = self.microsecond // 1000

        self.tzname: str = self.now.tzname()
        self.tzoffset: str = self.now.strftime("%z")

        self.date: str = f"{self.daynum_ofyear}/{self.year}"
        self.time: str = (
            f"{self.hour}:{self.minute}:{self.second}.{self.microsecond}"
        )

    def __repr__(self) -> str:
        r = (
            f"{self.year}, {self.month}, {self.day}, "
            f"{self.time}, {self.tzname}{self.tzoffset}"
        )
        return f"montyalex.MtaxTime({r})"

    def __str__(self) -> str:
        return self.timestamp()

    def __format__(self, format_spec: str) -> str:
        return self.timestamp(format_spec)

    def timestamp(self, format=None):
        if format is None:
            return f"{self.tzname}/{self.name}-{self.date}T{self.time}{self.tzoffset}"
        return datetime.strftime(self.now, format)



from montyalex.uo_tools._useroptions import UserOptions


# ----------------------------------------------------------------------
# |  YAML User Options Template
# ----------------------------------------------------------------------
class YamlOptions(UserOptions):
    def __init__(
        self, *, directory: str = ".mtax", filename: str = "settings"
    ) -> None:
        super().__init__(
            json=False,
            toml=False,
            yaml=True,
            mpck=False,
            directory=directory,
            filename=filename,
        )



# ----------------------------------------------------------------------
# |  Category
# ----------------------------------------------------------------------
from ._components import GroupSettingName as Group
from .subcategory import (
    SettingSubcategory as Subcategory,
    option_action_subcategory,
    info_default_subcategory,
    option_default_subcategory,
    locale_userinfo_subcategory,
)


# ----------------------------------------------------------------------
# |  Setting Category
# ----------------------------------------------------------------------
class SettingCategory:
    """
    A SettingCategory represents a category of settings within a
    settings container.

    It is defined by a prefix, type, and a variable number of
    subcategories.

    The prefix and infix are used to construct the ``info`` attribute,
    which is an instance of ``GroupSettingName``. This is used to
    identify the category.

    The type is a type that is used to check if a value is valid for
    the category.

    The subcategories are a tuple of instances of ``SettingSubcategory``.
    These are used to construct the settings within the category.

    The ``add_subcategory`` method is used to add additional subcategories
    to the category.

    The ``__str__`` method is used to get a string representation of the
    category, which is the string representation of the subcategories.

    The ``__repr__`` method is used to get a string representation of the
    category that can be evaluated as Python code.
    """

    def __init__(
        self, prefix: str, type_: type, *subcategories: Subcategory
    ) -> None:
        self.info: Group = Group(prefix)
        self.type_: type = type_
        self.subcategories: tuple[Subcategory] = subcategories

    def __str__(self) -> str:
        return f"{self.subcategories}"

    def __repr__(self) -> str:
        subcategories = f"{list(self.subcategories)!r}"
        return f"Category(={self.info.prefix}, type={self.type_}, subcategories={subcategories})"

    def add_subcategory(self, subcategory: Subcategory) -> None:
        new_subcategories = (*self.subcategories, subcategory)
        self.subcategories = new_subcategories


# ----------------------------------------------------------------------
# |  Categories
# ----------------------------------------------------------------------
ACTION = SettingCategory("action", object, option_action_subcategory)
DEFAULT = SettingCategory(
    "default",
    object,
    info_default_subcategory,
    option_default_subcategory,
)
USERINFO = SettingCategory("user", object, locale_userinfo_subcategory)



# ----------------------------------------------------------------------
# |  User-Options Tools
# ----------------------------------------------------------------------
from .json_schema_template import JsonSchemaOptions as schema
from .json_template import JsonOptions as json
from .mpck_template import MpckOptions as mpck
from .toml_template import TomlOptions as toml
from .yaml_template import YamlOptions as yaml


__version__ = "v1.0.0"


__all__ = ["schema", "json", "mpck", "toml", "yaml", "__version__"]



from montyalex.console_tools import richconsole
from montyalex.fs_tools.paths import rmfile
from montyalex.typing_tools import Any, Dict
from montyalex.uo_tools._template import UserTemplate

print = richconsole.print


# ----------------------------------------------------------------------
# |  Unpack items in a dictionary, "flatten"
# ----------------------------------------------------------------------
def unpack(
    d: Dict[str, Any], parent_key: str = "", sep: str = "."
) -> Dict[str, Any]:
    items: Dict[str, Any] = {}
    for k, v in d.items():
        new_key = f"{parent_key}{sep}{k}" if parent_key else k
        if isinstance(v, dict):
            items.update(unpack(v, new_key, sep=sep))
        else:
            items[new_key] = v
    return items


# ----------------------------------------------------------------------
# |  User Options
# ----------------------------------------------------------------------
class UserOptions(UserTemplate):
    def __init__(
        self,
        *,
        json: bool = False,
        toml: bool = False,
        yaml: bool = False,
        mpck: bool = False,
        directory: str = ".mtax",
        filename: str = "settings",
    ) -> None:
        super().__init__(json, toml, yaml, mpck, directory, filename)

    @property
    def options(self) -> Dict[str, Any]:
        return getattr(self, self.modelname)(read=True)

    @property
    def unpacked(self) -> Dict[str, Any]:
        return unpack

    @property
    def items(self):
        return self.options.items

    def remove(self):
        rmfile(self.modelpath)

    def change(
        self,
        key: str,
        value: Any,
        append: bool = True,
        overwrite: bool = False,
    ):
        d = {f"{key}": value}
        self.model(d, write=True, append=append, overwrite=overwrite)

    def inspect(
        self,
        mem_alloc: bool = False,
        representation: bool = False,
        full: bool = False,
        key: str = None,
    ):
        if key:
            third = None
            match self.modelname:
                case "json":
                    if ":" in key:
                        first, second = key.split(":", maxsplit=1)
                        if ":" in second:
                            third = key.split(":", maxsplit=1)[1]
                        if third:
                            print(
                                self.mpck(read=True)[f"{first}"][
                                    f"{second}"
                                ][f"{third}"]
                            )
                        else:
                            print(
                                self.json(read=True)[f"{first}"][
                                    f"{second}"
                                ]
                            )
                    else:
                        print(self.json(read=True)[f"{key}"])
                case "toml":
                    if ":" in key:
                        first, second = key.split(":", maxsplit=1)
                        if ":" in second:
                            third = key.split(":", maxsplit=1)[1]
                        if third:
                            print(
                                self.mpck(read=True)[f"{first}"][
                                    f"{second}"
                                ][f"{third}"]
                            )
                        else:
                            print(
                                self.toml(read=True)[f"{first}"][
                                    f"{second}"
                                ]
                            )
                    else:
                        print(self.toml(read=True)[f"{key}"])
                case "yaml":
                    if ":" in key:
                        first, second = key.split(":", maxsplit=1)
                        if ":" in second:
                            third = key.split(":", maxsplit=1)[1]
                        if third:
                            print(
                                self.mpck(read=True)[f"{first}"][
                                    f"{second}"
                                ][f"{third}"]
                            )
                        else:
                            print(
                                self.yaml(read=True)[f"{first}"][
                                    f"{second}"
                                ]
                            )
                    else:
                        print(self.yaml(read=True)[f"{key}"])
                case "mpck":
                    if ":" in key:
                        first, second = key.split(":", maxsplit=1)
                        if ":" in second:
                            third = key.split(":", maxsplit=1)[1]
                        if third:
                            print(
                                self.mpck(read=True)[f"{first}"][
                                    f"{second}"
                                ][f"{third}"]
                            )
                        else:
                            print(
                                self.mpck(read=True)[f"{first}"][
                                    f"{second}"
                                ]
                            )
                    else:
                        print(self.mpck(read=True)[f"{key}"])
        else:
            if self.exists:
                print(f"Found! [green]{self.modelpath}[/]")
                if full:
                    print("\n[green]User Options:[/]")
                if mem_alloc or full:
                    print(f"[bold dim]{self.options.keys}[/]")
                if representation or full:
                    print(f"[dim]{self.options.keys()}[/]")
                if mem_alloc or full:
                    print(f"[bold dim]{self.options.values}[/]")
                if representation or full:
                    print(f"[dim]{self.options.values()}[/]")
                if full:
                    print(f"{self.options}")
                print("\n[green]Iterated Items:[/]")
                if mem_alloc or full:
                    print(f"[bold dim]{self.items}[/]")
                print(f"[dim]{self.options}[/]")
                if representation or full:
                    print(f"[dim italic]{self.items()}[/]")
                for _k, _val in self.items():
                    print(f"{_k}: {_val}")
                print("\n[green]Unpacked Items:[/]")
                if mem_alloc or full:
                    print(f"[bold dim]{self.unpacked!r}[/]")
                print(f"[dim]{self.unpacked(self.options)}[/]")
                if representation or full:
                    print(
                        f"[dim italic]{self.unpacked(self.options).items()}[/]"
                    )
                for _k, _val in self.unpacked(self.options).items():
                    print(f"{_k}: {_val}")
                print("")
            else:
                print(f"Not Found! [red]{self.modelpath}[/]")


# ---------------------------------------------------------------------------
# |  Example Usage
# ---------------------------------------------------------------------------
if __name__ == "__main__":
    json_user_options = UserOptions(json=True)
    juo = json_user_options

    # ---------------------------------------------------------------------------
    # |  Run either A|B if there is not an initialized file with elements
    # ---------------------------------------------------------------------------
    # Run this if you would like to start from scratch or use overwrite
    # juo.remove()

    # (A)
    # d: Dict[str, Any] = {"": ""}  # Use this to add items
    # juo.model(d, write=True)  # Set overwrite=True if the file exists before running

    # (B)
    # juo.change(key='', value='')  # Or use this independently

    juo.inspect()



# ----------------------------------------------------------------------
# |  Directory Tools
# ----------------------------------------------------------------------
from .datedirs import (
    create_date_directories as datedirs,
    remove_date_directories as rmdatedirs,
)
from .simpledirs import (
    create_simple_directories as simpledirs,
    remove_simple_directories as rmsimpledirs,
)


__version__ = "v1.0.0"


__all__ = [
    "datedirs",
    "rmdatedirs",
    "simpledirs",
    "rmsimpledirs",
    "__version__",
]



# pylint: disable=line-too-long
# ----------------------------------------------------------------------
# |  _MontyAlex
# ----------------------------------------------------------------------
from montyalex.console_tools import warn_stm
from montyalex.fs_tools import joinpaths, mkdirs, current_working_dir


__version__ = "v0.1.4"


# ----------------------------------------------------------------------
# |  MontyAlex App Commands
# ----------------------------------------------------------------------
mtax_blueprints: list[
    tuple[str, str | None, int | None, str | None]
] = [
    ("├── mtax", "_mtax", "0", None),
    ("├── settings", "us_tools", "4", None),
    (
        "│ ├── init",
        "us_tools.[/][cyan italic]{uo_tools}[/][blue3 dim italic].(schema,json|toml|yaml|mpck).change()",
        "4",
        None,
    ),
    (
        "│ ├── list",
        "us_tools.[/][cyan italic]{uo_tools}[/][blue3 dim italic].(schema,json|toml|yaml|mpck).show_all_values()",
        "4",
        None,
    ),
    (
        "│ ├── delete",
        "us_tools.[/][cyan italic]{uo_tools}[/][blue3 dim italic].(schema,json|toml|yaml|mpck).remove()",
        "4",
        None,
    ),
    (
        "│ └── inspect",
        "us_tools.[/][cyan italic]{uo_tools}[/][blue3 dim italic].(schema,json|toml|yaml|mpck).inspect()",
        "4",
        None,
    ),
    ("├── mk", "directory_tools.single_directory", None, None),
    ("│ ├── [dim](-r/--range)[/]", None, None, None),
    (
        "│ ├── (-dirs/--simple-directories)",
        "directory_tools.simpledirs.create_simple_directories()",
        2,
        None,
    ),
    (
        "│ ├── (-datedirs/--date-directories)",
        "directory_tools.datedirs.create_date_directories()",
        3,
        None,
    ),
    ("│ ├── [dim](-4w/--four-weeks)[/]", None, None, None),
    ("│ ├── [dim](-p/--parent)[/]", None, None, None),
    ("│ ├── [dim](--name)[/]", None, None, None),
    ("│ ├── [dim](--prefix)[/]", None, None, None),
    ("│ ├── [dim](--suffix)[/]", None, None, None),
    ("│ └── [dim](-s)[/]", None, None, None),
    ("├── rm", "directory_tools.single_directory", None, None),
    (
        "│ ├── (-dirs/--simple-directories)",
        "directory_tools.simpledirs.remove_simple_directories()",
        4,
        None,
    ),
    (
        "│ └── (-datedirs/--date-directories)",
        "directory_tools.datedirs.remove_date_directories()",
        5,
        None,
    ),
    ("└── cache", "mtax_cache.MtaxCache", 6, None),
    ("  ├── add", "mtax_cache.MtaxCache.add_item()", None, None),
    ("  ├── clear", "mtax_cache.MtaxCache.clear()", None, None),
    ("  ├── get", "mtax_cache.MtaxCache.get_item()", None, None),
    ("  ├── info", "mtax_cache.MtaxCache.info()", None, None),
    ("  ├── list", "mtax_cache.MtaxCache.list_k_v_pairs()", None, None),
    ("  ├── remove", "mtax_cache.MtaxCache.remove()", None, None),
    ("  ├── reset", "mtax_cache.MtaxCache.reset()", None, None),
    ("  ├── set", "mtax_cache.MtaxCache.set_item()", None, None),
    (
        "  └── time",
        "mtax_cache.MtaxCache.[/][cyan italic]{MtaxTime}[/][blue3 dim italic]",
        None,
        None,
    ),
    ("    ├── reset", "mtax_cache.MtaxCache.timereset()", None, None),
    ("    └── set", "mtax_cache.MtaxCache.timeset()", None, None),
]

mtax_funcargs: list[tuple[str, str, int | str | None, str | None]] = [
    ("├── mtax", None, 6, None),
    (
        "│ ├── [green1 dim](-a)[/]",
        "Shows the arguments of functions",
        ".",
        None,
    ),
    (
        "│ ├── [green1 dim](-b)[/]",
        "Shows the blueprints ([default]classes[/]) of functions",
        ".",
        None,
    ),
    (
        "│ ├── [green1 dim](-c)[/]",
        "Shows the commands while showing args and/or blueprints",
        ".",
        None,
    ),
    (
        "│ ├── [green1 dim](-s)[/]",
        "Silence extraneous messages in console",
        ".",
        None,
    ),
    (
        "│ ├── [green1 dim](--save)[/]",
        "Save a file with the info here for projects or ease of use",
        ".",
        None,
    ),
    (
        "│ └── [green1 dim](--split)[/]",
        "Split the lines for better readability",
        ".",
        None,
    ),
    ("├── settings", None, None, None),
    ("│ ├── init", None, None, None),
    ("│ ├── list", None, None, None),
    ("│ ├── delete", None, None, None),
    ("│ └── inspect", None, None, None),
    ("├── mk", None, 7, None),
    ("│ ├── [green1 dim](-r/--range)[/]", None, ".", None),
    ("│ ├── [dim](-dirs)[/]", None, 6, None),
    ("│ ├── [dim](-datedirs)[/]", None, 3, None),
    ("│ ├── [green1 dim](-4w/--four-weeks)[/]", None, ".", None),
    ("│ ├── [green1 dim](-p/--parent)[/]", None, ".", None),
    ("│ ├── [green1 dim](--name)[/]", None, ".", None),
    ("│ ├── [green1 dim](--prefix)[/]", None, ".", None),
    ("│ ├── [green1 dim](--suffix)[/]", None, ".", None),
    (
        "│ └── [green1 dim](-s)[/]",
        "Silence extraneous messages in console",
        ".",
        None,
    ),
    ("├── rm", None, 7, None),
    ("│ ├── [green1 dim](-r/--range)[/]", None, ".", None),
    ("│ ├── [dim](-dirs)[/]", None, 6, None),
    ("│ ├── [dim](-datedirs)[/]", None, 3, None),
    ("│ ├── [green1 dim](-4w/--four-weeks)[/]", None, ".", None),
    ("│ ├── [green1 dim](-p/--parent)[/]", None, ".", None),
    ("│ ├── [green1 dim](--name)[/]", None, ".", None),
    ("│ ├── [green1 dim](--prefix)[/]", None, ".", None),
    ("│ ├── [green1 dim](--suffix)[/]", None, ".", None),
    (
        "│ └── [green1 dim](-s)[/]",
        "Silence extraneous messages in console",
        ".",
        None,
    ),
    ("└── cache", None, "0", None),
    ("  ├── add", None, 2, None),
    (
        "  │ ├── [green1 dim](--timezone)[/]",
        "Change the timezone object for time-data",
        ".",
        None,
    ),
    (
        "  │ └── [green1 dim](-s)[/]",
        "Silence extraneous messages in console",
        ".",
        None,
    ),
    ("  ├── clear", None, 2, None),
    (
        "  │ ├── [green1 dim](--timezone)[/]",
        "Change the timezone object for time-data",
        ".",
        None,
    ),
    (
        "  │ └── [green1 dim](-s)[/]",
        "Silence extraneous messages in console",
        ".",
        None,
    ),
    ("  ├── get", None, 2, None),
    (
        "  │ ├── [green1 dim](--timezone)[/]",
        "Change the timezone object for time-data",
        ".",
        None,
    ),
    (
        "  │ └── [green1 dim](-s)[/]",
        "Silence extraneous messages in console",
        ".",
        None,
    ),
    ("  ├── info", None, 2, None),
    (
        "  │ ├── [green1 dim](--timezone)[/]",
        "Change the timezone object for time-data",
        ".",
        None,
    ),
    (
        "  │ └── [green1 dim](-s)[/]",
        "Silence extraneous messages in console",
        ".",
        None,
    ),
    ("  ├── list", None, 2, None),
    (
        "  │ ├── [green1 dim](--timezone)[/]",
        "Change the timezone object for time-data",
        ".",
        None,
    ),
    (
        "  │ └── [green1 dim](-s)[/]",
        "Silence extraneous messages in console",
        ".",
        None,
    ),
    ("  ├── remove", None, 2, None),
    (
        "  │ ├── [green1 dim](--timezone)[/]",
        "Change the timezone object for time-data",
        ".",
        None,
    ),
    (
        "  │ └── [green1 dim](-s)[/]",
        "Silence extraneous messages in console",
        ".",
        None,
    ),
    ("  ├── reset", None, 2, None),
    (
        "  │ ├── [green1 dim](--timezone)[/]",
        "Change the timezone object for time-data",
        ".",
        None,
    ),
    (
        "  │ └── [green1 dim](-s)[/]",
        "Silence extraneous messages in console",
        ".",
        None,
    ),
    ("  ├── set", None, 2, None),
    (
        "  │ ├── [green1 dim](--timezone)[/]",
        "Change the timezone object for time-data",
        ".",
        None,
    ),
    (
        "  │ └── [green1 dim](-s)[/]",
        "Silence extraneous messages in console",
        ".",
        None,
    ),
    ("  └── time", None, "0", None),
    ("    ├── reset", None, 2, None),
    (
        "    │ ├── [green1 dim](--timezone)[/]",
        "Change the timezone object for time-data",
        ".",
        None,
    ),
    (
        "    │ └── [green1 dim](-s)[/]",
        "Silence extraneous messages in console",
        ".",
        None,
    ),
    ("    └── set", None, 2, None),
    (
        "      ├── [green1 dim](--timezone)[/]",
        "Change the timezone object for time-data",
        ".",
        None,
    ),
    (
        "      └── [green1 dim](-s)[/]",
        "Silence extraneous messages in console",
        ".",
        None,
    ),
]

mtax_commands: list[tuple[str, str, int | None, str | None]] = [
    ("├── mtax", "this command (prints list of commands)", 1, None),
    ("├── settings", "Operations for the settings of mtax", 2, None),
    ("│ ├── init", "Initializes a settings file", 3, None),
    ("│ ├── list", "Lists all mtax components available", 4, None),
    (
        "│ ├── delete",
        "Deletes the settings file or an item from the file",
        5,
        None,
    ),
    (
        "│ └── inspect",
        "Gives a report on info regarding settings in your env",
        6,
        None,
    ),
    (
        "├── mk",
        "Creates a single folder in the current local directory",
        None,
        None,
    ),
    ("│ ├── (-r)", "...", None, None),
    (
        "│ ├── (-dirs)",
        "Creates new blank folders in a range",
        7,
        "default_range: 1",
    ),
    (
        "│ ├── (-datedirs)",
        "Creates new datetime folders in a range",
        8,
        "default_range: *1yr",
    ),
    ("│ ├── (-4w)", "...", None, None),
    ("│ ├── (-p)", "...", None, None),
    ("│ ├── (--name)", "...", None, None),
    ("│ ├── (--prefix)", "...", None, None),
    ("│ ├── (--suffix)", "...", None, None),
    ("│ └── (-s)", "...", None, None),
    (
        "├── rm",
        "Removes a single folder in the current local directory",
        None,
        None,
    ),
    (
        "│ ├── (-dirs)",
        "Removes blank folders in a range",
        9,
        "default_range: 1",
    ),
    (
        "│ └── (-datedirs)",
        "Removes datetime folders in a range",
        10,
        "default_range: *1yr",
    ),
    ("└── cache", "Operations for the cache of mtax", None, None),
    (" ├── add", "Add to the cache", 11, None),
    (" ├── clear", "Clear the cache", 12, None),
    (" ├── get", "Gets an item from the cache", 13, None),
    (" ├── info", "Shows various info about the cache", 14, None),
    (" ├── list", "List the items stored in the cache", 15, None),
    (
        " ├── remove",
        "Remove the matching item from the cache",
        16,
        None,
    ),
    (" ├── reset", "Reset stuck items in the cache", 17, None),
    (" ├── set", "Set a item in the cache (Overwrite)", 18, None),
    (" └── time", "Manage time-data in the cache", None, None),
    ("    ├── reset", "Reset time-data in the cache", 19, None),
    (
        "    └── set",
        "Set a timestamp to time-data in the cache",
        20,
        None,
    ),
]


# ----------------------------------------------------------------------
# |  Save MontyAlex App Commands Markdown
# ----------------------------------------------------------------------
def mtax_complete(
    *,
    commands: bool = True,
    func_args: bool = False,
    blueprints: bool = False,
):
    mtax_available: list[str] = [
        "## `mtax`\n\n",
        "### `mtax --split`\n\n",
        "### `mtax --save`\n\n",
        "### `mtax -a`\n\n",
        "### `mtax -b`\n\n",
        "### `mtax -c`\n\n",
        "### `mtax -s`\n\n",
        "## `settings`\n\n",
        "### `settings init`\n\n",
        "### `settings list`\n\n",
        "### `settings delete`\n\n",
        "### `settings inspect`\n\n",
        "## `mk`\n\n",
        "### `mk -dirs`\n\n",
        "### `mk -datedirs`\n\n",
        "## `rm`\n\n",
        "### `rm -dirs`\n\n",
        "### `rm -datedirs`\n\n",
        "## `cache`\n\n",
        "## `cache add`\n\n",
        "## `cache clear`\n\n",
        "## `cache get`\n\n",
        "## `cache info`\n\n",
        "## `cache list`\n\n",
        "## `cache remove`\n\n",
        "## `cache reset`\n\n",
        "## `cache set`\n\n",
        "## `cache time`\n\n",
        "## `cache time reset`\n\n",
        "## `cache time set`\n\n",
    ]

    available: list[str] = [
        "The command that prints a list of commands (or generates this file)\n\n",
        "...\n\n",
        "...\n\n",
        "...\n\n",
        "...\n\n",
        "...\n\n",
        "...\n\n",
        "settings...\n\n",
        "init...\n\n",
        "list...\n\n",
        "delete...\n\n",
        "inspect...\n\n",
        "Creates a single folder in the current local directory\n\n",
        "Creates new blank folders in a range\n\n",
        "Creates new datetime folders in a range\n\n",
        "Removes a single folder in the current local directory\n\n",
        "Removes blank folders in a range\n\n",
        "Removes datetime folders in a range\n\n",
        "Operations for the cache of mtax\n\n",
        "Add to the cache\n\n",
        "Clear the cache\n\n",
        "Gets an item from the cache\n\n",
        "Shows various info about the cache\n\n",
        "List the items stored in the cache\n\n",
        "Remove the matching item from the cache\n\n",
        "Reset stuck items in the cache\n\n",
        "Set a item in the cache (Overwrite)\n\n",
        "Manage time-data in the cache\n\n",
        "Reset time-data in the cache\n\n",
        "Set a timestamp to time-data in the cache\n\n",
    ]
    templates: list[str] = [
        "  `...`\n\n",
        "  None\n\n",
        "  None\n\n",
        "  None\n\n",
        "  None\n\n",
        "  None\n\n",
        "  None\n\n",
        "  `...`\n\n",
        "  `...`\n\n",
        "  `...`\n\n",
        "  `...`\n\n",
        "  `...`\n\n",
        "  None\n\n",
        "  `...`\n\n",
        "  `...`\n\n",
        "  `...`\n\n",
        "  `...`\n\n",
        "  `...`\n\n",
        "  `...`\n\n",
        "  `...`\n\n",
        "  `...`\n\n",
        "  `...`\n\n",
        "  `...`\n\n",
        "  `...`\n\n",
        "  `...`\n\n",
        "  `...`\n\n",
        "  `...`\n\n",
        "  `...`\n\n",
        "  `...`\n\n",
        "  `...`\n\n",
        "  `...`\n\n",
        "  `...`\n\n",
        "  `...`\n\n",
    ]
    arguments: list[str] = [
        "- [--split]\n- [-s-a]\n- [-s-b]\n- [-s-c]\n- [--save]\n- [-a]\n- [-b]\n- [-c]\n- [-s]\n- [-ss]\n",
        "None\n",
        "None\n",
        "None\n",
        "None\n",
        "None\n",
        "None\n",
        "- ...\n",
        "- ...\n",
        "- ...\n",
        "- ...\n",
        "- ...\n",
        "None\n",
        "- ...\n",
        "- ...\n",
        "- ...\n",
        "- ...\n",
        "- ...\n",
        "- ...\n",
        "- ...\n",
        "- ...\n",
        "- ...\n",
        "- ...\n",
        "- ...\n",
        "- ...\n",
        "- ...\n",
        "- ...\n",
        "- ...\n",
        "- ...\n",
        "- ...\n",
    ]

    if commands:
        mtax_available = [
            item + desc for item, desc in zip(mtax_available, available)
        ]
    if blueprints:
        mtax_available = [
            item + desc for item, desc in zip(mtax_available, templates)
        ]
    if func_args:
        mtax_available = [
            item + desc for item, desc in zip(mtax_available, arguments)
        ]

    cwd = joinpaths(current_working_dir, ".mtax")
    cmd_fd = joinpaths(cwd, "commands.md")

    mkdirs(cwd, exist_ok=True)
    with open(cmd_fd, "w", encoding="UTF-8") as commands_file:
        commands_file.write(
            f"# MontyAlex\n\nA list of all mtax commands available as of {__version__}\n\n"
        )
        commands_file.writelines("\n".join(mtax_available))


def save_mtax_commands(
    *,
    blueprints: bool = False,
    commands: bool = False,
    func_args: bool = False,
    save_all: bool = False,
    save_commands: bool = False,
    save_funcargs: bool = False,
    save_blueprints: bool = False,
    silence: bool = False,
    super_silence: bool = False,
):
    """
    Save MontyAlex commands, function arguments, and blueprints to a markdown file.

    Args:
        blueprints (bool): Whether to include blueprints in the output.
        commands (bool): Whether to include commands in the output.
        func_args (bool): Whether to include function arguments in the output.
        save_all (bool): Save all information (commands, args, blueprints) to the file.
        save_commands (bool): Save only commands to the file.
        save_funcargs (bool): Save only function arguments to the file.
        save_blueprints (bool): Save only blueprints to the file.
        silence (bool): Suppress warnings and info messages.
        super_silence (bool): Suppress all console output.
    """
    if save_commands:
        mtax_complete(commands=True, func_args=False, blueprints=False)
    if save_funcargs:
        mtax_complete(commands=False, func_args=True, blueprints=False)
    if save_blueprints:
        mtax_complete(commands=False, func_args=False, blueprints=True)
    if save_all:
        mtax_complete(
            commands=commands,
            func_args=func_args,
            blueprints=blueprints,
        )
        if not silence and not super_silence:
            print(
                f"{warn_stm}, Syntax highlighting is not supported in files"
            )



# ----------------------------------------------------------------------
# |  MontyAlex App
# ----------------------------------------------------------------------
from ._mtax import save_mtax_commands
from ._settings import settings_
from .cache_tools import cache_, incr_cache_item, decr_cache_item

from .console_tools import richconsole
from .directory_tools import (
    datedirs,
    rmdatedirs,
    simpledirs,
    rmsimpledirs,
)
from .time_tools import func_time
from .typer_helpers import (
    blueprints_help,
    commands_help,
    date_directories,
    funcargs_help,
)
from .typer_tools import Option, Typer

print = richconsole.print


# ----------------------------------------------------------------------
# |  Typer App
# ----------------------------------------------------------------------
monty: Typer = Typer(
    name="monty",
    add_help_option=False,
    pretty_exceptions_show_locals=False,
)
monty.add_typer(settings_)
monty.add_typer(cache_)
LINEDASH = f'{"_" * 80}'


# ----------------------------------------------------------------------
# |  Typer Commands; create__dirs (mk), remove__dirs (rm), mtax
# ----------------------------------------------------------------------
@monty.command(name="mk")
@func_time("to create the directories")
def create__dirs(
    range_: int = Option(1, "--range", "-r"),
    s_dirs: bool = Option(
        False,
        "--simple-directories",
        "-dirs",
        is_flag=True,
        show_default="range: 1",
    ),
    dt_dirs: bool = Option(
        False,
        "--date-directories",
        "-datedirs",
        is_flag=True,
        show_default="range: 1yr",
    ),
    four_weeks: bool = Option(
        False, "--four-weeks", "-4w", is_flag=True
    ),
    dir_parent: str = Option(None, "--parent", "-p"),
    dir_name: str = Option(None, "--name"),
    prefix: str = Option(None, "--prefix"),
    suffix: str = Option(None, "--suffix"),
    silent: bool = Option(
        False,
        "-s",
        show_default="-!s",
        help="Silence extraneous messages in console",
    ),
) -> None:
    """
    Creates new folders in the current local directory

    Args:
        range_ (int): The number of directories to create.
        s_dirs (bool): A flag to create simple directories.
        dt_dirs (bool): A flag to create date directories.
        four_weeks (bool): A flag to specify four weeks.
        dir_parent (str): The parent directory.
        dir_name (str): The name of the directory.
        prefix (str): The prefix of the directory.
        suffix (str): The suffix of the directory.
        silent (bool): A flag to silence extraneous messages in console.
    """
    if s_dirs:
        simpledirs(range_, dir_name, prefix, suffix, silent)
        incr_cache_item(key="_mkdirs", silent=silent)
    if dt_dirs:
        date_directories(dir_name, prefix, suffix, silent)
        datedirs(range_, four_weeks, dir_parent, silent)
        incr_cache_item(key="_mkdates", silent=silent)
    if not s_dirs and not dt_dirs:
        print("Void, Please provide a type of directory to make")


@monty.command(name="rm")
@func_time("to remove the directories")
def remove__dirs(
    range_: int = Option(1, "--range", "-r"),
    s_dirs: bool = Option(
        False,
        "--simple-directories",
        "-dirs",
        is_flag=True,
        show_default="range: 1",
    ),
    dt_dirs: bool = Option(
        False,
        "--date-directories",
        "-datedirs",
        is_flag=True,
        show_default="range: 1yr",
    ),
    four_weeks: bool = Option(
        False, "--four-weeks", "-4w", is_flag=True
    ),
    dir_parent: str = Option(None, "--parent", "-p"),
    dir_name: str = Option(None, "--name", show_default="000"),
    prefix: str = Option(None, "--prefix"),
    suffix: str = Option(None, "--suffix"),
    silent: bool = Option(
        False,
        "-s",
        show_default="-!s",
        help="Silence extraneous messages in console",
    ),
):
    """
    Removes folders in the current local directory

    Usage:
        >>> mtax rm -dirs -name 001 -prefix prefix- -suffix -suffix
        >>> mtax rm -datedirs -name 001 -prefix prefix- -suffix -suffix
        >>> mtax rm -dirs -name 001 -prefix prefix- -suffix -suffix -silent
        >>> mtax rm -datedirs -name 001 -prefix prefix- -suffix -suffix -silent
    """
    if s_dirs:
        rmsimpledirs(range_, dir_name, prefix, suffix, silent)
        decr_cache_item(key="_mkdirs", silent=silent)
    if dt_dirs:
        date_directories(dir_name, prefix, suffix, silent)
        rmdatedirs(range_, four_weeks, dir_parent, silent)
        decr_cache_item(key="_mkdates", silent=silent)
    else:
        print("Void, Please provide a type of directory to remove")


@monty.command()
@monty.command(name="help")
@cache_.command(name="help")
@settings_.command(name="help")
def mtax(
    help_: bool = Option(False, "--help", "-h"),
    split: bool = Option(False, "--split"),
    save_funcargs_to_file: bool = Option(False, "--save-a", "-s-a"),
    save_blueprints_to_file: bool = Option(False, "--save-b", "-s-b"),
    save_commands_to_file: bool = Option(False, "--save-c", "-s-c"),
    save_all: bool = Option(False, "--save"),
    func_args: bool = Option(False, "-a", "--args"),
    blueprints: bool = Option(False, "-b", "--blueprints"),
    commands: bool = Option(False, "-c", "--commands"),
    super_silence: bool = Option(False, "-ss", show_default="-!ss"),
    silence: bool = Option(False, "-s", show_default="-!s"),
):
    """Prints the list of MontyAlex commands, their arguments, and their blueprints

    Args:
        help: bool = Option(False, "--help", "-h")
            Print help for the command.
        split: bool = Option(False, "--split")
            Split the lines for better readability.
        save_funcargs_to_file: bool = Option(False, "--save-a", "-s-a")
            Save the arguments of functions to a file.
        save_blueprints_to_file: bool = Option(False, "--save-b", "-s-b")
            Save the blueprints (classes) of functions to a file.
        save_commands_to_file: bool = Option(False, "--save-c", "-s-c")
            Save the commands to a file.
        save_all: bool = Option(False, "--save")
            Save all to a file.
        func_args: bool = Option(False, "-a", "--args")
            Show the arguments of functions.
        blueprints: bool = Option(False, "-b", "--blueprints")
            Show the blueprints (classes) of functions.
        commands: bool = Option(False, "-c", "--commands")
            Show the commands.
        super_silence: bool = Option(False, "-ss", show_default="-!ss")
            Silence all messages in console.
        silence: bool = Option(False, "-s", show_default="-!s")
            Silence extraneous messages in console.
    """
    if help_:
        print("HELP")
    else:
        if not super_silence:
            print(LINEDASH)
            commands_help(
                split, (not blueprints and not func_args) or commands
            )
            if (func_args or blueprints) and commands:
                print(LINEDASH)
            funcargs_help(split, func_args)
            if func_args and blueprints:
                print(LINEDASH)
            blueprints_help(split, blueprints)
            print(LINEDASH)
        save_mtax_commands(
            blueprints=blueprints,
            commands=commands,
            func_args=func_args,
            save_all=save_all,
            save_blueprints=save_blueprints_to_file,
            save_commands=save_commands_to_file,
            save_funcargs=save_funcargs_to_file,
            silence=silence,
            super_silence=super_silence,
        )



default_settings_template = {
    "action.dirs.datetime": {
        "default.opt.four-week-month": False,
        "default.opt.format": "%Y-%m-%d",
        "default.opt.range": "1yr",
    },
    "action.dirs.simple": {"default.opt.range": 1},
    "action.opt.silent": False,
    "default.info.timezone": None,
    "user.locale.city": None,
    "user.locale.continent": None,
    "user.locale.country": None,
    "user.locale.state": None,
    "user.locale.street-name": None,
    "user.locale.street-number": None,
}

empty_settings_template = {
    "action.dirs.datetime": {
        "default.opt.four-week-month": False,
        "default.opt.range": "1yr",
    },
    "action.dirs.simple": {"default.opt.range": 1},
    "action.opt.silent": False,
    "default.info.timezone": "",
    "user.locale.city": "",
    "user.locale.continent": "",
    "user.locale.country": "",
    "user.locale.state": "",
    "user.locale.street-name": "",
    "user.locale.street-number": "",
}

empty_settings_plus_fmt = {
    "action.dirs.datetime": {
        "default.opt.four-week-month": False,
        "default.opt.format": "%Y-%m-%d",
        "default.opt.range": "1yr",
    },
    "action.dirs.simple": {"default.opt.range": 1},
    "action.opt.silent": False,
    "default.info.timezone": "",
    "user.locale.city": "",
    "user.locale.continent": "",
    "user.locale.country": "",
    "user.locale.state": "",
    "user.locale.street-name": "",
    "user.locale.street-number": "",
}

settings_schema_definition = {
    "description": "",
    "type": "object",
    "properties": {
        "action.dirs.datetime": {
            "description": "",
            "type": "object",
            "properties": {
                "default.opt.four-week-month": {
                    "description": "Use a four week (28d) range instead of a complete month",
                    "type": ["boolean", "null"],
                    "default": False,
                },
                "default.opt.format": {
                    "description": "Change the date format used for folders",
                    "type": ["string", "null"],
                    "default": "%Y-%m-%d",
                },
                "default.opt.range": {
                    "description": "",
                    "type": ["string", "null"],
                    "default": "1yr",
                },
            },
        },
        "action.dirs.simple": {
            "description": "",
            "type": "object",
            "properties": {
                "default.opt.range": {
                    "description": "",
                    "type": ["integer", "null"],
                    "default": 1,
                }
            },
        },
        "action.opt.silent": {
            "description": "",
            "type": ["boolean", "null"],
            "default": False,
        },
        "default.info.timezone": {
            "description": "Default timezone used for date related commands",
            "type": ["string", "null"],
        },
    },
}



# ----------------------------------------------------------------------
# |  _Defaults
# ----------------------------------------------------------------------
from ._components import (
    GroupSettingName,
    ShortSettingName,
    SettingName,
    SettingComponent,
)
from ._actions import datetime_directory_action, simple_directory_action


# ----------------------------------------------------------------------
# |  Default Setting Component
# ----------------------------------------------------------------------
class Default(SettingComponent):
    """
    A Default setting component represents a default setting.

    It is defined by an infix, suffix, type, and optionally a prefix and parent.

    The prefix and infix are used to construct the setting's identifier.
    The parent is an instance of a setting name class that provides hierarchical context.
    The type is used to validate associated values.
    """

    def __init__(
        self,
        infix: str,
        suffix: str,
        type_: type,
        *,
        prefix: str = "default",
        parent: (
            GroupSettingName | ShortSettingName | SettingName
        ) = None,
    ) -> None:
        super().__init__(
            prefix, infix, suffix, type_=type_, parent=parent
        )


# ----------------------------------------------------------------------
# |  Default Instances
# ----------------------------------------------------------------------
timezone_info_default = Default(
    "info",
    "timezone",
    str,
    parent=SettingName("default", "info", "timezone"),
)
timeout_option_default = Default(
    "opt",
    "timeout",
    str,
    parent=SettingName("default", "opt", "timeout"),
)
fourweek_option_default = Default(
    "opt", "four-week-month", bool, parent=datetime_directory_action
)
datetime_range_option_default = Default(
    "opt", "range", str, parent=datetime_directory_action
)
datetime_format_option_default = Default(
    "opt", "format", str, parent=datetime_directory_action
)
simple_range_option_default = Default(
    "opt", "range", str, parent=simple_directory_action
)



# ----------------------------------------------------------------------
# |  Tree App
# ----------------------------------------------------------------------
from montyalex.console_tools import richconsole
from montyalex.typer_tools import Option, Typer
from .presets import LPR, GPR, MCR, Table
from .treeid import FileSystemTree

print = richconsole.print


# ----------------------------------------------------------------------
# |  Typer App
# ----------------------------------------------------------------------
tree_: Typer = Typer(name="tree", pretty_exceptions_show_locals=False)
nodes_: Typer = Typer(name="nodes", pretty_exceptions_show_locals=False)
tree_.add_typer(nodes_)


# ----------------------------------------------------------------------
# |  Typer Commands; create_, setup_, first_,
# |  next_, prev_, last_, table_, delete_
# ----------------------------------------------------------------------
@tree_.command(name="create")
def create_(id_: str):
    fs_tree = FileSystemTree(id_)
    if fs_tree.exists:
        print(f"Tree {id_!r} already exists")
    else:
        fs_tree.write()
        print(f"Tree created at: {id_!r}, use setup to continue")


@tree_.command(name="setup")
def setup_(id_: str, template: str = None):
    fs_tree = FileSystemTree(id_)
    scanned_tree = FileSystemTree(id_).scan()
    if template:
        fs_tree.setup(template=template)
    fs_tree_name = scanned_tree["name"]
    fs_tree_age = scanned_tree["age"]
    fs_tree_template = None
    if "template" in scanned_tree.keys():
        fs_tree_template = scanned_tree["template"]
    print(f"Tree: {fs_tree_name!r}, {fs_tree_age!r}")
    if fs_tree_template:
        print(f"Template: {fs_tree_template!r}")


@nodes_.command(name="first")
def first_(id_: str):
    fs_tree = FileSystemTree(id_).scan()
    fs_nodes = fs_tree["nodes"]
    first_node = fs_nodes[0]
    print(f'First item of {fs_tree["name"]!r} tree: {first_node}')


@nodes_.command(name="next")
def next_(id_: str):
    fs_tree = FileSystemTree(id_).scan()
    print(f'Next item of {fs_tree["name"]!r} tree: ')


@nodes_.command(name="prev")
def prev_(id_: str):
    fs_tree = FileSystemTree(id_).scan()
    print(f'Prev item of {fs_tree["name"]!r} tree: ')


@nodes_.command(name="last")
def last_(id_: str):
    fs_tree = FileSystemTree(id_).scan()
    fs_nodes = fs_tree["nodes"]
    last_node = fs_nodes[-1]
    print(f'Last item of {fs_tree["name"]!r} tree: {last_node}')


@tree_.command(name="table")
def table_(
    least: bool = Option(False, "-LPR", "--least-range"),
    most_common: bool = Option(False, "-MCR", "--common-range"),
    greatest: bool = Option(False, "-GPR", "--greatest-range"),
    show_wz_range: bool = Option(False, "-Z", "--show-zero-range"),
):
    if least:
        tables: dict[str, Table] = {
            k: v.table(show_wz_range) for k, v in LPR.items()
        }
        for table in tables.values():
            table.title = "L.P. " + table.title
            print(table)
    if greatest:
        tables: dict[str, Table] = {
            k: v.table(show_wz_range) for k, v in GPR.items()
        }
        for table in tables.values():
            table.title = "G.P. " + table.title
            print(table)
    if most_common:
        tables: dict[str, Table] = {
            k: v.table(show_wz_range) for k, v in MCR.items()
        }
        for table in tables.values():
            table.title = "M.C. " + table.title
            print(table)
    if show_wz_range:
        print(
            "The zeroith directory for the digit amount specified is included"
        )
    else:
        print(
            "The zeroith directory for the digit amount specified is reserved"
        )


@tree_.command(name="delete")
def delete_(id_: str):
    fs_tree = FileSystemTree(id_)
    if fs_tree.exists is True:
        print(f"Tree {id_!r} deleted")
        fs_tree.delete()
    else:
        print(f"Tree {id_!r} already does not exist")



# ----------------------------------------------------------------------
# |  Subcategory
# ----------------------------------------------------------------------
from ._components import (
    ShortSettingName as Short,
    SettingComponent as Component,
)
from ._actions import silent_option_action
from ._defaults import (
    datetime_range_option_default,
    datetime_format_option_default,
    timeout_option_default,
    fourweek_option_default,
    simple_range_option_default,
    timezone_info_default,
)
from ._userinfo import (
    locale_city_userinfo,
    locale_continent_userinfo,
    locale_country_userinfo,
    locale_province_userinfo,
    locale_region_userinfo,
    locale_state_userinfo,
    locale_streetname_userinfo,
    locale_streetnum_userinfo,
)


# ----------------------------------------------------------------------
# |  Setting Subcategory
# ----------------------------------------------------------------------
class SettingSubcategory:
    """
    A SettingSubcategory represents a subcategory of settings within a
    category.

    It is defined by a prefix, infix, type, and a variable number of
    components.

    The prefix and infix are used to construct the ``info`` attribute,
    which is an instance of ``ShortSettingName``. This is used to
    identify the subcategory.

    The type is a type that is used to check if a value is valid for
    the subcategory.

    The components are a tuple of instances of ``SettingComponent``.
    These are used to construct the settings within the subcategory.

    The ``add_component`` method is used to add additional components to
    the subcategory.

    The ``__str__`` method is used to get a string representation of the
    subcategory, which is the string representation of the components.

    The ``__repr__`` method is used to get a string representation of the
    subcategory that can be evaluated as Python code.
    """

    def __init__(
        self,
        prefix: str,
        infix: str,
        type_: type,
        *components: Component,
    ) -> None:
        self.info: Short = Short(prefix, infix)
        self.type_: type = type_
        self.components: tuple[Component] = components

    def __str__(self) -> str:
        return f"{self.components}"

    def __repr__(self) -> str:
        return (
            f"Subcategory(={self.info.prefix}.{self.info.infix}, "
            f"type={self.type_}, components={list(self.components)!r})"
        )

    def add_component(self, component: Component) -> None:
        new_components = (*self.components, component)
        self.components = new_components


# ----------------------------------------------------------------------
# |  Subcategories
# ----------------------------------------------------------------------
option_action_subcategory = SettingSubcategory(
    "action", "opt", object, silent_option_action
)
info_default_subcategory = SettingSubcategory(
    "default", "info", object, timezone_info_default
)
option_default_subcategory = SettingSubcategory(
    "default",
    "opt",
    object,
    timeout_option_default,
    fourweek_option_default,
    datetime_range_option_default,
    datetime_format_option_default,
    simple_range_option_default,
)
locale_userinfo_subcategory = SettingSubcategory(
    "user",
    "locale",
    object,
    locale_region_userinfo,
    locale_continent_userinfo,
    locale_country_userinfo,
    locale_state_userinfo,
    locale_province_userinfo,
    locale_city_userinfo,
    locale_streetname_userinfo,
    locale_streetnum_userinfo,
)



from montyalex.typer_tools import Option
from montyalex.uo_tools import json, toml, yaml, mpck
from montyalex.us_tools import SETTINGS
from .initialization import initialize_void


def inspect_useroptions(
    dirname: str = Option(".mtax", "--directory-name", "-dir"),
    filename: str = Option("settings", "--file-name", "-name"),
    *,
    use_all: bool = Option(False, "--all", "-a"),
    use_synced: bool = Option(False, "--sync", "-s"),
    use_json: bool = Option(False, "--json", "-j"),
    use_toml: bool = Option(False, "--toml", "-t"),
    use_yaml: bool = Option(False, "--yaml", "-y"),
    use_mpck: bool = Option(False, "--mpck", "-m"),
    full_inspection: bool = Option(False, "--full", "-fi"),
    mem_alloc_inspection: bool = Option(False, "--memory", "-mi"),
    repr_inspection: bool = Option(False, "--repr", "-ri"),
    exists_inspection: bool = Option(False, "--exists", "-ei"),
    key_inspection: str = Option(None, "--key", "-ki"),
):
    uo: json | toml | yaml | mpck | None = None
    if use_json:
        juo: json = json(directory=dirname, filename=filename)
        uo = juo
    if use_toml:
        tuo: toml = toml(directory=dirname, filename=filename)
        uo = tuo
    if use_yaml:
        yuo: yaml = yaml(directory=dirname, filename=filename)
        uo = yuo
    if use_mpck:
        muo: mpck = mpck(directory=dirname, filename=filename)
        uo = muo
    if uo is None:
        uo = json(directory=dirname, filename=filename)

    inspection: bool = False
    initialize_void(
        uo,
        dirname,
        filename,
        use_all=use_all,
        use_synced=use_synced,
        use_json=use_json,
        use_toml=use_toml,
        use_yaml=use_yaml,
        use_mpck=use_mpck,
    )
    if full_inspection:
        inspection = True
        uo.inspect(full=True)
        print("[green]Verified Settings:[/]")
        print(f"{SETTINGS!r}")
        SETTINGS.show_all_values()
        # if repr_listing:
        #     print(f'{SETTINGS!r}')
        # else:
        #     print(f'{SETTINGS!r}')
        #     SETTINGS.show_all_values()
    if mem_alloc_inspection:
        inspection = True
        uo.inspect(mem_alloc=True)
    if repr_inspection:
        inspection = True
        uo.inspect(representation=True)
    if key_inspection:
        inspection = True
        uo.inspect(key=key_inspection)
    if exists_inspection:
        inspection = True
        if uo.exists:
            print(f"Found! [green]{uo.modelpath}[/]")
        else:
            print(f"Not Found! [red]{uo.modelpath}[/]")
    if not inspection:
        uo.inspect()



from montyalex._mtax import (
    mtax_blueprints,
    mtax_commands,
    mtax_funcargs,
)
from montyalex.typing_tools import NoneType


def commands_help(split: bool = False, commands: bool = False):
    if commands:
        if split:
            print("")
        print("cmds.")
        if split:
            print("")
        for name, shorthelp, index, defaults in mtax_commands:
            if isinstance(index, int):
                if index <= 9:
                    helpmsg = f"([pink3 dim]{index}[/]) {name} -> [dim italic]{shorthelp}[/]"
                else:
                    helpmsg = f"([pink3 dim]{index}[/]){name} -> [dim italic]{shorthelp}[/]"
            if isinstance(index, NoneType):
                helpmsg = (
                    f"([dim]X[/]) {name} -> [dim italic]{shorthelp}[/]"
                )
            if defaults:
                helpmsg += f" [pink3 dim italic]({defaults})[/]"
            print(helpmsg)
            if split:
                print("")


def funcargs_help(split: bool = False, func_args: bool = False):
    if func_args:
        if split:
            print("")
        print("args.")
        if split:
            print("")
        for (
            func,
            module,
            funcargs_index,
            funcargs_defaults,
        ) in mtax_funcargs:
            if funcargs_index:
                if funcargs_index != ".":
                    helpmsg = f"([green1 dim]{funcargs_index}[/]) [italic]{func}[/]"
                else:
                    helpmsg = (
                        f"([dim]{funcargs_index}[/]) [italic]{func}[/]"
                    )
            else:
                helpmsg = f"    [italic]{func}[/]"
            if module:
                helpmsg += f" -> [green1 dim italic]{module}[/]"
            if funcargs_defaults:
                helpmsg += f" [dim italic]({funcargs_defaults})[/]"
            if not module and not funcargs_defaults:
                helpmsg += " [dim italic]...[/]"
            print(helpmsg)
            if split:
                print("")


def blueprints_help(split: bool = False, blueprints: bool = False):
    if blueprints:
        if split:
            print("")
        print("bpns.")
        if split:
            print("")
        for (
            blueprint,
            module,
            blueprint_index,
            blueprint_defaults,
        ) in mtax_blueprints:
            if blueprint_index:
                helpmsg = (
                    f"([blue3 dim]{blueprint_index}[/]) {blueprint}"
                )
            else:
                helpmsg = f"    {blueprint}"
            if module:
                helpmsg += f" -> [blue3 dim italic]{module}[/]"
            if blueprint_defaults:
                helpmsg += f" [dim italic]({blueprint_defaults})[/]"
            if not module and not blueprint_defaults:
                helpmsg += " [dim italic]...[/]"
            print(helpmsg)
            if split:
                print("")



from montyalex.typer_tools import Option
from montyalex.uo_tools import json, toml, yaml, mpck


def delete_useroptions(
    dirname: str = Option(".mtax", "--directory-name", "-dir"),
    filename: str = Option("settings", "--file-name", "-name"),
    key: str = Option(None, "--key", "-k"),
    *,
    use_json: bool = Option(False, "--json", "-j"),
    use_toml: bool = Option(False, "--toml", "-t"),
    use_yaml: bool = Option(False, "--yaml", "-y"),
    use_mpck: bool = Option(False, "--mpck", "-m"),
):
    uo: json | toml | yaml | mpck | None = None
    if use_json:
        uo = json(directory=dirname, filename=filename)
    if use_toml:
        uo = toml(directory=dirname, filename=filename)
    if use_yaml:
        uo = yaml(directory=dirname, filename=filename)
    if use_mpck:
        uo = mpck(directory=dirname, filename=filename)
    if uo is None:
        uo = json(directory=dirname, filename=filename)

    if key:
        uo.change(key, None, overwrite=True)
    uo.remove()



# ----------------------------------------------------------------------
# |  (Time)Zone-Cache App
# ----------------------------------------------------------------------
from montyalex.time_tools import MtaxTime
from montyalex.typer_tools import Option, Typer
from .mtax_cache import MtaxCache


# ----------------------------------------------------------------------
# |  Typer App
# ----------------------------------------------------------------------
zone_: Typer = Typer(
    name="zone", help="Manage timezone-data in the cache"
)


# ----------------------------------------------------------------------
# |  Typer Commands; reset_, set_, get_
# ----------------------------------------------------------------------
@zone_.command(name="reset")
def reset_(
    *,
    timezone: str = "Etc/Greenwich",
    silent: bool = Option(
        False,
        "-s",
        show_default="-!s",
        help="Silence extraneous messages in console",
    ),
):
    """Reset timezone in the cache"""
    mtax: MtaxCache = MtaxCache(timezone)
    mtax_time = MtaxTime(timezone)
    mtax.remove("_timezone_", silent)
    set_(timezone=mtax_time.timezone, silent=silent)


@zone_.command(name="set")
def set_(
    *,
    timezone: str = "Etc/Greenwich",
    silent: bool = Option(
        False,
        "-s",
        show_default="-!s",
        help="Silence extraneous messages in console",
    ),
):
    """Set timezone in the cache"""
    mtax: MtaxCache = MtaxCache(timezone)
    mtax_time = MtaxTime(timezone)
    mtax.set_item(("_timezone_", mtax_time.timezone), silent)


@zone_.command(name="get")
def get_(
    *,
    silent: bool = Option(
        False,
        "-s",
        show_default="-!s",
        help="Silence extraneous messages in console",
    ),
):
    """Set timezone in the cache"""
    mtax: MtaxCache = MtaxCache("Etc/Greenwich")
    mtax.get_item("_timezone_", silent)



# ----------------------------------------------------------------------
# |  _Actions
# ----------------------------------------------------------------------
from ._components import (
    GroupSettingName,
    ShortSettingName,
    SettingName,
    SettingComponent,
)


# ----------------------------------------------------------------------
# |  Action Setting Component
# ----------------------------------------------------------------------
class Action(SettingComponent):
    """An Action setting component represents a specific action-related setting.

    It is defined by an infix, suffix, type, and optionally a prefix and parent.

    The prefix, infix, and suffix are used to construct the setting's identifier.
    The parent is an instance of a setting name class that provides hierarchical context.
    The type is used to validate associated values.
    """

    def __init__(
        self,
        infix: str,
        suffix: str,
        type_: type,
        *,
        prefix: str = "action",
        parent: (
            GroupSettingName | ShortSettingName | SettingName
        ) = None,
    ) -> None:
        super().__init__(
            prefix, infix, suffix, type_=type_, parent=parent
        )


# ----------------------------------------------------------------------
# |  Action Instances
# ----------------------------------------------------------------------
datetime_directory_action = Action(
    "dirs",
    "datetime",
    object,
    parent=SettingName("action", "dirs", "datetime"),
)
simple_directory_action = Action(
    "dirs",
    "simple",
    object,
    parent=SettingName("action", "dirs", "simple"),
)
silent_option_action = Action(
    "opt", "silent", bool, parent=SettingName("action", "opt", "silent")
)



# ----------------------------------------------------------------------
# |  Void
# ----------------------------------------------------------------------
from collections import Counter
from montyalex.fs_tools import joinpaths, current_working_dir


# ----------------------------------------------------------------------
# |  Find, Warn & Return Duplicates in a Non-Unique List
# ----------------------------------------------------------------------
def find_duplicates(nonunique_list: list[str], obj: object):
    counts = Counter(nonunique_list)
    duplicates = {
        item: count for item, count in counts.items() if count > 1
    }

    if duplicates:
        print(f"{obj} name duplicates found:")
        for item, count in duplicates.items():
            print(f"'{item}' is repeated {count} times")

    return duplicates


# ----------------------------------------------------------------------
# |  Named Empty Files
# ----------------------------------------------------------------------
class NamedEmptyFile:
    def __init__(self, name: str) -> None:
        self.file_name: str = name

    def __str__(self) -> str:
        return "File"


# ----------------------------------------------------------------------
# |  Named Directories
# ----------------------------------------------------------------------
class NamedDirectory:
    def __init__(self, name: str) -> None:
        self.directory_name: str = name

    def __str__(self) -> str:
        return "Directory"


# ----------------------------------------------------------------------
# |  Base for File System Trees
# ----------------------------------------------------------------------
class Tree:
    def __init__(
        self,
        name: str,
        directories: set[NamedDirectory],
        files: set[NamedEmptyFile],
    ) -> None:
        self.name: str = name
        self.named_directories: set[NamedDirectory] = directories
        self.named_files: set[NamedEmptyFile] = files


# ----------------------------------------------------------------------
# |  Void File System Tree Template
# ----------------------------------------------------------------------
class VoidTree:
    def __init__(self, name: str) -> None:
        self.name = name
        self.os_path = joinpaths(current_working_dir, self.name)
        self.directory_set: set[NamedDirectory] = set()
        self.file_set: set[NamedEmptyFile] = set()
        self.tree = Tree(self.name, self.directory_set, self.file_set)

    def _add_directory_set(self, directory_set: set[NamedDirectory]):
        self.directory_set = directory_set
        self.tree.named_directories = self.directory_set

    def _add_file_set(self, file_set: set[NamedEmptyFile]):
        self.file_set = file_set
        self.tree.named_files = self.file_set

    def add_directories(self, directories: list[NamedDirectory]):
        find_duplicates(directories, NamedDirectory(""))
        self._add_directory_set(set(directories))

    def add_files(self, files: list[NamedEmptyFile]):
        find_duplicates(files, NamedEmptyFile(""))
        self._add_file_set(set(files))

    def get_tree(self):
        return self.tree.name

    def get_directories(self) -> set:
        return self.directory_set

    def get_files(self) -> set:
        return self.file_set

    def display_tree(self):
        tree = self.get_tree()
        print(tree)

    def display_directories(self):
        directories = self.get_directories()
        if directories:
            print(f"{directories!r}")

    def display_files(self):
        files = self.get_files()
        if files:
            print(f"{files!r}")


# ---------------------------------------------------------------------------
# |  Example Usage
# ---------------------------------------------------------------------------
if __name__ == "__main__":
    dirs_ = [
        # "dir1",
        # "dir2",
        # "dir2",
        # "dir2",
        # "dir2",
        # "dir2",
        # "dir2",
        # "dir3",
    ]

    files_ = [
        # "file1",
        # "file2",
        # "file2",
        # "file2",
        # "file2",
        # "file2",
        # "file2",
        # "file2",
        # "file4",
        # "file3",
        # "file3",
    ]

    void = VoidTree("name")

    void.add_directories(dirs_)
    void.add_files(files_)

    void.display_tree()
    void.display_directories()
    void.display_files()



# ----------------------------------------------------------------------
# |  Date-Directories
# ----------------------------------------------------------------------
from datetime import datetime, timedelta
from dateutil.relativedelta import relativedelta

from montyalex.cache_tools import add_cache_item, remove_cache_item
from montyalex.console_tools import richconsole, success_stm
from montyalex.fs_tools import (
    abspath,
    current_working_dir,
    joinpaths,
    mkdirs,
    pathexists,
)
from montyalex.us_tools import datetime_format_option_default
from .helpdirs import remove_formatted__dir

print = richconsole.print


FORMAT = datetime_format_option_default.get_value()
if FORMAT == "Notfound":
    FORMAT = "%Y-%m-%d"
START = datetime.now()


def create_date_directories(
    range_: int = 1,
    four_weeks: bool = False,
    parent: str = None,
    silent: bool = False,
) -> None:
    add_cache_item(
        key="_datedirs_", value=f"{START:FORMAT}", silent=True
    )
    new_date = START
    for i in range(range_ * 12):
        if four_weeks:
            new_date += timedelta(days=28)
        else:
            new_date = START + relativedelta(months=i)
        formatted_date = new_date.strftime(FORMAT)
        if parent:
            mkdirs(
                joinpaths(current_working_dir, parent, formatted_date),
                exist_ok=True,
            )
        else:
            mkdirs(
                joinpaths(current_working_dir, formatted_date),
                exist_ok=True,
            )
        if not silent:
            cwd_ = abspath(parent) if parent else current_working_dir
            print(
                f"{success_stm}, Created {formatted_date!r} in {cwd_!r}"
            )


def remove_date_directories(
    range_: int = 1,
    four_weeks: bool = False,
    parent: str = None,
    silent: bool = False,
):
    remove_cache_item(key="_datedirs_", silent=True)
    fmsg = f"{success_stm}, Removed "
    if parent:
        remove_formatted__dir(parent, silent)

        new_date = START
        for i in range(range_ * 12):
            if four_weeks:
                new_date += timedelta(days=28)
            else:
                new_date = START + relativedelta(months=i)
            formatted_date = new_date.strftime(FORMAT)
            formatted_directory = joinpaths(
                current_working_dir, formatted_date
            )
            formatted_directory = joinpaths(current_working_dir, parent)
            if not pathexists(formatted_directory) and not silent:
                print(
                    f"{fmsg}{formatted_date!r} from {formatted_directory!r}"
                )
        if not pathexists(parent) and not silent:
            print(f"{fmsg}{parent!r} from {current_working_dir!r}")
    else:
        for i in range(range_ * 12):
            new_date = START + relativedelta(months=i)
            formatted_date = new_date.strftime(FORMAT)
            formatted_directory = joinpaths(
                current_working_dir, formatted_date
            )
            remove_formatted__dir(formatted_directory, silent)
            if not pathexists(formatted_directory) and not silent:
                print(
                    f"{fmsg}{formatted_date!r} from {current_working_dir!r}"
                )



# ----------------------------------------------------------------------
# |  Exceptions
# ----------------------------------------------------------------------
from json import JSONDecodeError
from montyalex.typing_tools import Any


# ----------------------------------------------------------------------
# |  Base Exceptions
# ----------------------------------------------------------------------
class MontyBaseException(BaseException):
    def __init__(
        self,
        message: str,
        value: Any = None,
        name: str = "MontyBaseException",
    ) -> None:
        super().__init__(message)
        self.name = name
        self.message = message
        self.value = value

    def __str__(self) -> str:
        return f"{self.name}: {self.message} {self.value}"

    def __repr__(self) -> str:
        return self.name

    def __reduce__(self) -> tuple[Any, tuple[str, Any]]:
        return self.__class__, (self.message, self.value)


class MontyArithmeticException(ArithmeticError):
    def __init__(
        self,
        message: str,
        value: Any = None,
        name: str = "MontyArithmeticException",
    ) -> None:
        super().__init__(message)
        self.name = name
        self.message = message
        self.value = value

    def __str__(self) -> str:
        return f"{self.name}: {self.message} {self.value}"

    def __repr__(self) -> str:
        return self.name

    def __reduce__(self) -> tuple[Any, tuple[str, Any]]:
        return self.__class__, (self.message, self.value)


class MontyJSONException(JSONDecodeError):
    def __init__(
        self,
        message: str,
        value: Any = None,
        name: str = "MontyJSONException",
    ) -> None:
        super().__init__(message, "", 0)
        self.name = name
        self.message = message
        self.value = value

    def __str__(self) -> str:
        return f"{self.name}: {self.message} {self.value}"

    def __repr__(self) -> str:
        return self.name

    def __reduce__(self) -> tuple[Any, tuple[str, Any]]:
        return self.__class__, (self.message, self.value)


class MontyAlexException(MontyBaseException):
    def __init__(
        self, message: str, value: Any, name: str = "MontyAlexException"
    ) -> None:
        super().__init__(message, value, name)


class MontyAppException(MontyAlexException):
    def __init__(
        self, message: str, value: Any, name: str = "MontyAppException"
    ) -> None:
        super().__init__(message, value, name)


class MontyTyperException(MontyBaseException):
    def __init__(
        self,
        message: str,
        value: Any,
        name: str = "MontyTyperException",
    ) -> None:
        super().__init__(message, value, name)


class MontyOSException(MontyAppException):
    def __init__(
        self, message: str, value: Any, name: str = "MontyOSException"
    ) -> None:
        super().__init__(message, value, name)


class MontyUserException(MontyAppException):
    def __init__(
        self, message: str, value: Any, name: str = "MontyUserException"
    ) -> None:
        super().__init__(message, value, name)


# ----------------------------------------------------------------------
# |  Custom Errors
# ----------------------------------------------------------------------
class SettingsError(MontyAppException):
    def __init__(
        self,
        message: str,
        value: Any = None,
        name: str = "SettingsError",
    ) -> None:
        super().__init__(message, value, name)


class TyperAppError(MontyTyperException):
    def __init__(
        self,
        message: str,
        value: Any = None,
        name: str = "TyperAppError",
    ) -> None:
        super().__init__(message, value, name)


class CacheError(MontyAppException):
    def __init__(
        self, message: str, value: Any = None, name: str = "CacheError"
    ) -> None:
        super().__init__(message, value, name)


class ConsoleError(MontyAppException):
    def __init__(
        self,
        message: str,
        value: Any = None,
        name: str = "ConsoleError",
    ) -> None:
        super().__init__(message, value, name)


class DirectoryError(MontyOSException):
    def __init__(
        self,
        message: str,
        value: Any = None,
        name: str = "DirectoryError",
    ) -> None:
        super().__init__(message, value, name)


class FileSystemError(MontyOSException):
    def __init__(
        self,
        message: str,
        value: Any = None,
        name: str = "FileSystemError",
    ) -> None:
        super().__init__(message, value, name)


class TimeError(MontyAppException):
    def __init__(
        self, message: str, value: Any = None, name: str = "TimeError"
    ) -> None:
        super().__init__(message, value, name)


class UserOptionsError(MontyUserException):
    def __init__(
        self,
        message: str,
        value: Any = None,
        name: str = "UserOptionsError",
    ) -> None:
        super().__init__(message, value, name)


class UserSettingsError(MontyUserException):
    def __init__(
        self,
        message: str,
        value: Any = None,
        name: str = "UserSettingsError",
    ) -> None:
        super().__init__(message, value, name)


class RangeError(MontyArithmeticException):
    def __init__(
        self, message: str, value: Any = None, name: str = "RangeError"
    ) -> None:
        super().__init__(message, value, name)


class KeyValueError(MontyBaseException):
    def __init__(
        self, message: str, value: Any = None, name: str = "RangeError"
    ) -> None:
        super().__init__(message, value, name)


class JSONError(MontyJSONException):
    def __init__(
        self, message: str, value: Any = None, name: str = "JSONError"
    ) -> None:
        super().__init__(message, value, name)



# ----------------------------------------------------------------------
# |  Help-Directories
# ----------------------------------------------------------------------
import shutil

from montyalex.console_tools import richconsole, error_stm
from montyalex.fs_tools import pathexists

print = richconsole.print


def remove_formatted__dir(
    dir: str | bytes, silent: bool = False
) -> bool:
    if pathexists(dir):
        try:
            shutil.rmtree(dir)
            return True
        except (OSError, NotADirectoryError, IsADirectoryError):
            if not silent:
                print(
                    f"{error_stm}, Removing {dir!r} failed in the directory given"
                )
            return False
    else:
        if not silent:
            print(
                f"Warning, Attempted to remove {dir!r}, "
                + "but it didn't exist"
            )
        return False



# ----------------------------------------------------------------------
# |  _Settings
# ----------------------------------------------------------------------
from .console_tools import richconsole
from .typer_helpers import (
    delete_useroptions,
    initialize_useroptions,
    inspect_useroptions,
)
from .typer_tools import Option, Typer
from .us_tools import SETTINGS

print = richconsole.print


# ----------------------------------------------------------------------
# |  Typer App
# ----------------------------------------------------------------------
settings_: Typer = Typer(
    name="settings",
    add_help_option=False,
    pretty_exceptions_show_locals=False,
)


# ----------------------------------------------------------------------
# |  Typer Commands; init_, list_, delete_, inspect_
# ----------------------------------------------------------------------
@settings_.command(name="list", add_help_option=False)
def list_(
    repr_listing: bool = Option(False, "--repr", "-r"),
):
    """List the current settings

    Args:
        repr_listing (bool): If True, print the repr representation of SETTINGS;
                             otherwise, show all values.
    """
    if repr_listing:
        print(f"{SETTINGS!r}")
    else:
        SETTINGS.show_all_values()


@settings_.command(name="init", add_help_option=False)
@settings_.command(name="init", add_help_option=False)
def init_(
    dirname: str = Option(".mtax", "--directory-name", "-dir"),
    filename: str = Option("settings", "--file-name", "-name"),
    *,
    overwrite: bool = Option(False, "--overwrite", "-o"),
    use_all: bool = Option(False, "--all", "-a"),
    use_synced: bool = Option(False, "--sync", "-s"),
    use_json: bool = Option(False, "--json", "-j"),
    use_toml: bool = Option(False, "--toml", "-t"),
    use_yaml: bool = Option(False, "--yaml", "-y"),
    use_mpck: bool = Option(False, "--mpck", "-m"),
):
    """Initialize settings file

    Args:
        dirname (str): The directory to put the settings file in.
        filename (str): The name of the settings file.
        overwrite (bool): If True, overwrite the existing file.
        use_all (bool): If True, use all formats.
        use_synced (bool): If True, use the synced format.
        use_json (bool): If True, use the JSON format.
        use_toml (bool): If True, use the TOML format.
        use_yaml (bool): If True, use the YAML format.
        use_mpck (bool): If True, use the MPCK format.
    """
    initialize_useroptions(
        dirname=dirname,
        filename=filename,
        overwrite=overwrite,
        use_all=use_all,
        use_synced=use_synced,
        use_json=use_json,
        use_toml=use_toml,
        use_yaml=use_yaml,
        use_mpck=use_mpck,
    )


@settings_.command(name="delete", add_help_option=False)
def delete_(
    dirname: str = Option(".mtax", "--directory-name", "-dir"),
    filename: str = Option("settings", "--file-name", "-name"),
    key: str = Option(None, "--key", "-k"),
    *,
    use_json: bool = Option(False, "--json", "-j"),
    use_toml: bool = Option(False, "--toml", "-t"),
    use_yaml: bool = Option(False, "--yaml", "-y"),
    use_mpck: bool = Option(False, "--mpck", "-m"),
):
    """Delete settings file or a key from settings file

    Args:
        dirname (str): The directory to put the settings file in.
        filename (str): The name of the settings file.
        key (str): The key to delete from the settings file.
        use_json (bool): If True, use the JSON format.
        use_toml (bool): If True, use the TOML format.
        use_yaml (bool): If True, use the YAML format.
        use_mpck (bool): If True, use the MPCK format.
    """
    delete_useroptions(
        dirname=dirname,
        filename=filename,
        key=key,
        use_json=use_json,
        use_toml=use_toml,
        use_yaml=use_yaml,
        use_mpck=use_mpck,
    )


@settings_.command(name="inspect", add_help_option=False)
def inspect_(
    dirname: str = Option(".mtax", "--directory-name", "-dir"),
    filename: str = Option("settings", "--file-name", "-name"),
    use_all: bool = Option(False, "--all", "-a"),
    use_synced: bool = Option(False, "--sync", "-s"),
    use_json: bool = Option(False, "--json", "-j"),
    use_toml: bool = Option(False, "--toml", "-t"),
    use_yaml: bool = Option(False, "--yaml", "-y"),
    use_mpck: bool = Option(False, "--mpck", "-m"),
    full_inspection: bool = Option(False, "--full", "-fi"),
    mem_alloc_inspection: bool = Option(False, "--memory", "-mi"),
    repr_inspection: bool = Option(False, "--repr", "-ri"),
    exists_inspection: bool = Option(False, "--exists", "-ei"),
    key_inspection: str = Option(None, "--key", "-ki"),
) -> None:
    """Inspect settings file

    Args:
        dirname (str): The directory to put the settings file in.
        filename (str): The name of the settings file.
        use_all (bool): If True, use all formats.
        use_synced (bool): If True, use the synced format.
        use_json (bool): If True, use the JSON format.
        use_toml (bool): If True, use the TOML format.
        use_yaml (bool): If True, use the YAML format.
        use_mpck (bool): If True, use the MPCK format.
        full_inspection (bool): If True, inspect the entire file.
        mem_alloc_inspection (bool): If True, inspect the memory allocation.
        repr_inspection (bool): If True, inspect the representation.
        exists_inspection (bool): If True, inspect if the file exists.
        key_inspection (str): The key to inspect.
    """
    inspect_useroptions(
        dirname,
        filename,
        use_all=use_all,
        use_synced=use_synced,
        use_json=use_json,
        use_toml=use_toml,
        use_yaml=use_yaml,
        use_mpck=use_mpck,
        full_inspection=full_inspection,
        mem_alloc_inspection=mem_alloc_inspection,
        repr_inspection=repr_inspection,
        exists_inspection=exists_inspection,
        key_inspection=key_inspection,
    )



# ----------------------------------------------------------------------
# |  Paths
# ----------------------------------------------------------------------
import os
from posixpath import abspath, expanduser
import shutil
from pathlib import Path


# ----------------------------------------------------------------------
# |  Global shorthand for os, path, and file system functions
# ----------------------------------------------------------------------
joinpaths: callable = os.path.join
pathexists: callable = os.path.exists
_: callable = abspath, expanduser
current_working_dir: str = os.getcwd()
current_working_path: str = Path().cwd()


def rmfile(
    path,
):
    os.remove(path=path)


def mkdirs(name, mode: int = 511, exist_ok: bool = False):
    os.makedirs(name=name, mode=mode, exist_ok=exist_ok)


def rmdirs(
    name,
):
    os.removedirs(name=name)


def rmtree(
    path,
):
    shutil.rmtree(path=path)



# ----------------------------------------------------------------------
# |  User-Settings
# ----------------------------------------------------------------------
from montyalex.console_tools import richconsole
from montyalex.typing_tools import Any
from .category import (
    SettingCategory as Category,
    ACTION,
    DEFAULT,
    USERINFO,
)

print = richconsole.print


# ----------------------------------------------------------------------
# |  Settings
# ----------------------------------------------------------------------
class Settings:
    def __init__(self) -> None:
        self.action: Category = ACTION
        self.default: Category = DEFAULT
        self.userinfo: Category = USERINFO

        self.all: list[Category] = [
            self.action,
            self.default,
            self.userinfo,
        ]

    def __str__(self) -> str:
        return f"{self.all}"

    def __repr__(self) -> str:
        return f"Settings({self.all})"

    def get_value(self, key: str, value_only: bool = True) -> Any:
        for category in self.all:
            for subcat in category.subcategories:
                for comp in subcat.components:
                    if f"{comp}" == key:
                        value = comp.get_value()
                        if value_only:
                            return value
                        if f"{comp.parent}" != f"{comp}":
                            return f"mtax.{comp.parent}.{comp}: {value}"
                        return f"mtax.{comp}: {value}"
        return None

    def show_value(self, key: str) -> None:
        for category in self.all:
            for subcat in category.subcategories:
                for comp in subcat.components:
                    if f"{comp}" == key:
                        value = comp.get_value()
                        if f"{comp.parent}" != f"{comp}":
                            print(f"mtax.{comp.parent}.{comp}: {value}")
                        else:
                            print(f"mtax.{comp}: {value}")

    def show_all_values(self) -> None:
        components: list[Category] = []
        for category in self.all:
            for subcat in category.subcategories:
                for component in subcat.components:
                    components.append(component)
        components.sort(
            key=lambda c: (
                c.parent.prefix.value,
                c.parent.infix.value,
                c.parent.suffix.value,
            )
        )
        for comp in components:
            value = comp.get_value()
            type_ = f"{comp.type_}".removeprefix(
                "<class "
            ).removesuffix(">")
            if f"{comp.parent}" != f"{comp}":
                if value != "Notfound":
                    print(
                        f"[yellow italic]mtax.[/][orange1]{comp.parent}[/]"
                        f".[yellow]{comp}[/]: {value!r}"
                    )
                else:
                    print(
                        f"[yellow italic]mtax.[/][orange1]{comp.parent}[/]"
                        f".[yellow]{comp}[/]: [red]{value}[/]"
                        f"[dim] (Value should be of type {type_})[/]"
                    )
            else:
                if value != "Notfound":
                    print(
                        f"[yellow italic]mtax.[/][yellow]{comp}[/]: {value!r}"
                    )
                else:
                    print(
                        f"[yellow italic]mtax.[/][yellow]{comp}[/]: [red]{value}[/]"
                        f"[dim] (Value should be of type {type_})[/]"
                    )


# ----------------------------------------------------------------------
# |  Global Settings Object
# ----------------------------------------------------------------------
SETTINGS = Settings()



from abc import abstractmethod
from functools import lru_cache
import msgspec

from montyalex.fs_tools import (
    current_working_dir,
    joinpaths,
    pathexists,
    mkdirs,
)
from montyalex.typing_tools import Any, Dict, Optional


# ----------------------------------------------------------------------
# |  User Template (json, toml, yaml, mpck)
# ----------------------------------------------------------------------
class UserTemplate:
    def __init__(
        self,
        json: bool = False,
        toml: bool = False,
        yaml: bool = False,
        mpck: bool = False,
        directory: str = ".mtax",
        filename: str = "settings",
    ) -> None:
        self._cwd: str = current_working_dir

        self.dirname: str = directory
        self.filename: str = filename

        self.dirpath: str = joinpaths(self._cwd, self.dirname)
        self.filepath: str = joinpaths(self.dirpath, self.filename)

        self._dir_exists: str = pathexists(self.dirpath)
        self._file_exists: str = pathexists(self.filepath)

        selected_formats = [
            fmt
            for fmt, enabled in {
                "json": json,
                "toml": toml,
                "yaml": yaml,
                "mpck": mpck,
            }.items()
            if enabled
        ]

        if not selected_formats:
            raise ValueError(
                "At least one format must be enabled (json, toml, yaml, mpck)."
            )

        if len(selected_formats) > 1:
            raise ValueError(
                "Only one format can be enabled at a time."
            )

        selected_format = selected_formats[0]

        self.modelname = selected_format
        self.modelpath = f"{self.filepath}.{self.modelname}"
        match self.modelname:
            case "json":
                self.model = self.json
            case "toml":
                self.model = self.toml
            case "yaml":
                self.model = self.yaml
            case "mpck":
                self.model = self.mpck

    @property
    def exists(self) -> bool:
        return pathexists(self.modelpath)

    @property
    @abstractmethod
    def options(self) -> Dict[str, Any]:
        raise NotImplementedError

    @property
    @abstractmethod
    def unpacked(self) -> Dict[str, Any]:
        raise NotImplementedError

    @property
    @abstractmethod
    def items(self) -> Dict[str, Any]:
        raise NotImplementedError

    @abstractmethod
    def inspect(self):
        raise NotImplementedError

    @lru_cache(maxsize=16)
    def __read_settings(self) -> Dict[str, Any]:
        if pathexists(self.modelpath):
            with open(self.modelpath, "rb") as settings_file:
                data = settings_file.read()
            if self.modelname == "json":
                return self._djson(data)
            if self.modelname == "toml":
                return self._dtoml(data)
            if self.modelname == "yaml":
                return self._dyaml(data)
            if self.modelname == "mpck":
                return self._dmpck(data)
        return {}

    def __save_settings(self, encoded_data: bytes) -> None:
        if not self._dir_exists:
            mkdirs(self.dirpath)
        with open(self.modelpath, "wb") as settings_file:
            settings_file.write(encoded_data)

    def convert(self, *args, **kwargs):
        msgspec.convert(*args, **kwargs)

    def json(
        self,
        data: Optional[Any] = None,
        read: bool = False,
        write: bool = False,
        append: bool = False,
        overwrite: bool = False,
    ) -> Dict[str, Any] | None:
        if read:
            return self.__read_settings()
        if write and data is not None:
            self._ejson(data, append=append, overwrite=overwrite)
        return None

    def _ejson(
        self,
        data: Dict[str, Any],
        append: bool = False,
        overwrite: bool = False,
    ) -> None:
        if append:
            existing = self.__read_settings()
            existing.update(data)
            encoded = msgspec.json.encode(existing)
        elif overwrite or not self.exists:
            encoded = msgspec.json.encode(data)
        else:
            encoded = msgspec.json.encode(data)
        encoded = msgspec.json.format(encoded, indent=4)
        self.__save_settings(encoded)

    def _djson(self, *args, **kwargs):
        return msgspec.json.decode(*args, **kwargs)

    def toml(
        self,
        data: Optional[Any] = None,
        read: bool = False,
        write: bool = False,
        append: bool = False,
        overwrite: bool = False,
    ) -> Dict[str, Any] | None:
        if read:
            return self.__read_settings()
        if write and data is not None:
            self._etoml(data, append=append, overwrite=overwrite)
        return None

    def _etoml(
        self,
        data: Dict[str, Any],
        append: bool = False,
        overwrite: bool = False,
    ) -> None:
        if append:
            existing = self.__read_settings()
            existing.update(data)
            encoded = msgspec.toml.encode(existing)
        elif overwrite or not self.exists:
            encoded = msgspec.toml.encode(data)
        else:
            encoded = msgspec.toml.encode(data)
        self.__save_settings(encoded)

    def _dtoml(self, *args, **kwargs):
        return msgspec.toml.decode(*args, **kwargs)

    def yaml(
        self,
        data: Optional[Any] = None,
        read: bool = False,
        write: bool = False,
        append: bool = False,
        overwrite: bool = False,
    ) -> Dict[str, Any] | None:
        if read:
            return self.__read_settings()
        if write and data is not None:
            self._eyaml(data, append=append, overwrite=overwrite)
        return None

    def _eyaml(
        self,
        data: Dict[str, Any],
        append: bool = False,
        overwrite: bool = False,
    ) -> None:
        if append:
            existing = self.__read_settings()
            existing.update(data)
            encoded = msgspec.yaml.encode(existing)
        elif overwrite or not self.exists:
            encoded = msgspec.yaml.encode(data)
        else:
            encoded = msgspec.yaml.encode(data)
        self.__save_settings(encoded)

    def _dyaml(self, *args, **kwargs):
        return msgspec.yaml.decode(*args, **kwargs)

    def mpck(
        self,
        data: Optional[Any] = None,
        read: bool = False,
        write: bool = False,
        append: bool = False,
        overwrite: bool = False,
    ) -> Dict[str, Any] | None:
        if read:
            return self.__read_settings()
        if write and data is not None:
            self._empck(data, append=append, overwrite=overwrite)
        return None

    def _empck(
        self,
        data: Dict[str, Any],
        append: bool = False,
        overwrite: bool = False,
    ) -> None:
        if append:
            existing = self.__read_settings()
            existing.update(data)
            encoded = msgspec.msgpack.encode(existing)
        elif overwrite or not self.exists:
            encoded = msgspec.msgpack.encode(data)
        else:
            encoded = msgspec.msgpack.encode(data)
        self.__save_settings(encoded)

    def _dmpck(self, *args, **kwargs):
        return msgspec.msgpack.decode(*args, **kwargs)


# ----------------------------------------------------------------------
# |  Example Usage
# ----------------------------------------------------------------------
if __name__ == "__main__":
    standard_user_template = UserTemplate(json=True)
    std_temp = standard_user_template

    d: Dict[str, Any] = {"": ""}
    std_temp.model(d, write=True)



# ----------------------------------------------------------------------
# |  Decorators
# ----------------------------------------------------------------------
from montyalex.fs_tools import cancel
from montyalex.object_tools import Key, Value, RangeError
from montyalex.typing_tools import Any, Callable, Literal
from montyalex.uo_tools import yaml


# ----------------------------------------------------------------------
# |  State-Manager
# ----------------------------------------------------------------------
def statemanager_decorator(
    key: Key, value: Value, file_name: str, *, loc: str = None
):
    """
    A decorator that saves the given key-value pair in a YAML file
    and prints the state before calling the function.

    Args:
        key (Key): A key for the state.
        value (Value): The value for the key.
        file_name (str): The name of the YAML file.
        loc (str, optional): The location of the YAML file. Defaults to ".mtax".

    Returns:
        Callable[[Callable], Callable]: A decorator.
    """
    yaml_state = yaml(
        directory=loc if loc else ".mtax", filename=file_name
    )

    def statemanager(func):
        state: dict[Key, Any] = {key: value}

        def wrapper():
            print(f"{state}")
            func()

        yaml_state.model(state, write=True, overwrite=True)
        return wrapper

    print(yaml_state.modelpath)
    return statemanager


# ----------------------------------------------------------------------
# |  State-Modifier Helper Function for Incr/Decr
# ----------------------------------------------------------------------
def statemodifier(
    value: Value,
    operation: Literal[0] | Literal[1],
    multiplier: int = 1,
    limit: int | tuple[int, int] = None,
    ilimit: int | tuple[int, int] = None,
) -> bool:
    """
    A helper function for state modification. It checks if the given value
    is within the given limit or inclusive limit after the given operation.
    If it is not, it raises a RangeError.

    Args:
        value (Value): The value to check.
        operation (Literal[0] | Literal[1]): The operation to perform.
            0 for increment, 1 for decrement.
        multiplier (int, optional): The multiplier for the operation.
            Defaults to 1.
        limit (int | tuple[int, int], optional): The limit for the value.
            Defaults to None.
        ilimit (int | tuple[int, int], optional): The inclusive limit for the value.
            Defaults to None.

    Raises:
        RangeError: If the value is not within the given limit or inclusive limit.

    Returns:
        bool: True if the value is within the given limit or inclusive limit.
    """
    if isinstance(limit, tuple) or isinstance(ilimit, tuple):
        range_limit = (
            range(*limit)
            if not ilimit
            else range(ilimit[0], ilimit[1] + 1)
        )
        check = (
            value + multiplier if operation == 0 else value - multiplier
        )
        if value not in range_limit or check not in range_limit:
            if limit:
                raise RangeError(
                    f"Index ({check}) not in", range(*limit)
                )
            if ilimit:
                raise RangeError(
                    f"Index ({check}) not in", range(*ilimit)
                )
            cancel()
    if isinstance(limit, int) or isinstance(ilimit, int):
        range_limit = range(limit) if not ilimit else range(ilimit + 1)
        check = (
            value + multiplier if operation == 0 else value - multiplier
        )
        if value not in range_limit or check not in range_limit:
            if limit:
                raise RangeError(
                    f"Index ({check}) not in", range(limit)
                )
            if ilimit:
                raise RangeError(
                    f"Index ({check}) not in", range(ilimit)
                )
            cancel()
    return True


# ----------------------------------------------------------------------
# |  State-Decr (-multiplier)
# ----------------------------------------------------------------------
def statedecr_decorator(
    key: Key,
    file_name: str,
    *,
    multiplier: int = 1,
    limit: int | tuple[int, int] = None,
    ilimit: int | tuple[int, int] = None,
    loc: str = None,
):
    """
    Decorator to decrement a value in a settings file.

    Args:
        key (Key): The key to decrement.
        file_name (str): The name of the settings file.
        multiplier (int, optional): The multiplier for the decrement. Defaults to 1.
        limit (int | tuple[int, int], optional): The limit for the decrement. Defaults to None.
        ilimit (int | tuple[int, int], optional): The inclusive limit for the decrement. Defaults to None.
        loc (str, optional): The location of the settings file. Defaults to None.

    Returns:
        Callable[[Callable], Callable]: The decorated function.
    """
    yaml_state = yaml(
        directory=loc if loc else ".mtax", filename=file_name
    )

    def statedecrementer(func):
        key_value = yaml_state.model(read=True)[key]
        state: dict[Key, Any] = {}
        if isinstance(key_value, int):
            state: dict[Key, Any] = {key: key_value - multiplier}

        def wrapper():
            func()

        try:
            smdecr = statemodifier(
                key_value, 1, multiplier, limit, ilimit
            )
            if state[key] != key_value and smdecr:
                print(f"{state}")
                yaml_state.model(state, write=True, overwrite=True)
                print(yaml_state.modelpath)
        except RangeError as e:
            print(e)
        return wrapper

    return statedecrementer


# ----------------------------------------------------------------------
# |  State-Incr (+multiplier)
# ----------------------------------------------------------------------
def stateincr_decorator(
    key: Key,
    file_name: str,
    *,
    multiplier: int = 1,
    limit: int | tuple[int, int] = None,
    ilimit: int | tuple[int, int] = None,
    loc: str = None,
) -> Callable[[Callable], Callable]:
    """
    Decorator to increment a value in a settings file.

    Args:
        key (Key): The key to increment.
        file_name (str): The name of the settings file.
        multiplier (int, optional): The multiplier for the increment. Defaults to 1.
        limit (int | tuple[int, int], optional): The limit for the increment. Defaults to None.
        ilimit (int | tuple[int, int], optional): The inclusive limit for the increment. Defaults to None.
        loc (str, optional): The location of the settings file. Defaults to None.

    Returns:
        Callable[[Callable], Callable]: The decorated function.
    """
    yaml_state = yaml(
        directory=loc if loc else ".mtax", filename=file_name
    )

    def stateincrementer(func):
        key_value = yaml_state.model(read=True)[key]
        state: dict[Key, Any] = {}
        if isinstance(key_value, int):
            state: dict[Key, Any] = {key: key_value + multiplier}

        def wrapper():
            func()

        try:
            smincr = statemodifier(
                key_value, 0, multiplier, limit, ilimit
            )
            if state[key] != key_value and smincr:
                print(f"{state}")
                yaml_state.model(state, write=True, overwrite=True)
                print(yaml_state.modelpath)
        except RangeError as e:
            print(e)
        return wrapper

    return stateincrementer



# ----------------------------------------------------------------------
# |  _Components
# ----------------------------------------------------------------------
from abc import abstractmethod
from functools import total_ordering
from json import JSONDecodeError

from montyalex.fs_tools import (
    pathexists,
    joinpaths,
    current_working_dir,
)
from montyalex.typing_tools import Any
from montyalex.uo_tools import json, toml, yaml, mpck


# ----------------------------------------------------------------------
# |  _Prefix, _Infix, _Suffix (SettingName)
# ----------------------------------------------------------------------
class _Prefix:
    def __init__(self, value: str) -> None:
        self.value = value

    def __str__(self) -> str:
        return self.value

    def __repr__(self) -> str:
        return f"Prefix({self.value})"


class _Infix:
    def __init__(self, value: str) -> None:
        self.value = value

    def __str__(self) -> str:
        return self.value

    def __repr__(self) -> str:
        return f"Infix({self.value})"


class _Suffix:
    def __init__(self, value: str) -> None:
        self.value = value

    def __str__(self) -> str:
        return self.value

    def __repr__(self) -> str:
        return f"Suffix({self.value})"


# ----------------------------------------------------------------------
# |  GroupSettingName, ShortSettingName, SettingName (User-Setting)
# ----------------------------------------------------------------------
@total_ordering
class GroupSettingName:
    def __init__(self, prefix: str) -> None:
        self.prefix: _Prefix = _Prefix(prefix)

    @abstractmethod
    def __eq__(self, other: object) -> bool:
        if not isinstance(other, SettingName):
            return NotImplemented
        return self.prefix.value == other.prefix.value

    @abstractmethod
    def __lt__(self, other: object) -> bool:
        if not isinstance(other, SettingName):
            return NotImplemented
        if self.prefix.value < other.prefix.value:
            return True
        return False

    @abstractmethod
    def __hash__(self):
        return hash((self.prefix.value))


@total_ordering
class ShortSettingName(GroupSettingName):
    def __init__(self, prefix: str, infix: str) -> None:
        super().__init__(prefix)
        self.infix: _Infix = _Infix(infix)

    @abstractmethod
    def __eq__(self, other: object) -> bool:
        if not isinstance(other, SettingName):
            return NotImplemented
        return (self.prefix.value, self.infix.value) == (
            other.prefix.value,
            other.infix.value,
        )

    @abstractmethod
    def __lt__(self, other: object) -> bool:
        if not isinstance(other, SettingName):
            return NotImplemented
        if self.prefix.value < other.prefix.value:
            return True
        if self.prefix.value > other.prefix.value:
            return False
        return self.infix.value < other.infix.value

    @abstractmethod
    def __hash__(self):
        return hash((self.prefix.value, self.infix.value))


class SettingName(ShortSettingName):
    def __init__(self, prefix: str, infix: str, suffix: str) -> None:
        super().__init__(prefix, infix)
        self.suffix: _Suffix = _Suffix(suffix)

    def __str__(self) -> str:
        return f"{self.prefix}.{self.infix}.{self.suffix}"

    def __repr__(self) -> str:
        return f"Name({self.prefix}.{self.infix}.{self.suffix})"

    @abstractmethod
    def __eq__(self, other: object) -> bool:
        if not isinstance(other, SettingName):
            return NotImplemented
        return (self.prefix.value, self.infix.value) == (
            other.prefix.value,
            other.infix.value,
        )

    @abstractmethod
    def __lt__(self, other: object) -> bool:
        if not isinstance(other, SettingName):
            return NotImplemented
        if self.prefix.value < other.prefix.value:
            return True
        if self.prefix.value > other.prefix.value:
            return False
        if self.infix.value < other.infix.value:
            return True
        if self.infix.value > other.infix.value:
            return False
        return self.suffix.value < other.suffix.value

    @abstractmethod
    def __hash__(self):
        return hash(
            (self.prefix.value, self.infix.value, self.suffix.value)
        )


# ----------------------------------------------------------------------
# |  SettingComponent
# ----------------------------------------------------------------------
class SettingComponent(SettingName):
    """
    A SettingComponent represents a specific setting within a user's configuration.

    It is defined by a prefix, infix, suffix, and optionally a parent and type.

    The prefix, infix, and suffix are used to construct the setting's identifier.
    The parent is an instance of a setting name class that provides hierarchical context.
    The type is used to validate associated values.

    Methods:
    - __str__: Returns the string representation of the setting's identifier.
    - __repr__: Returns a detailed string representation of the setting.
    - __eq__: Compares two settings for equality based on their identifiers.
    - __lt__: Compares two settings to determine their order based on their identifiers.
    - __hash__: Provides a hash representation of the setting.
    - get_value: Retrieves the setting's value from the user's configuration files.
    """

    def __init__(
        self,
        prefix: str,
        infix: str,
        suffix: str,
        *,
        parent: (
            GroupSettingName | ShortSettingName | SettingName
        ) = None,
        type_: type,
    ) -> None:
        super().__init__(prefix, infix, suffix)
        self.parent: (
            GroupSettingName | ShortSettingName | SettingName
        ) = parent
        self.type_: type = type_
        self.value: str | bool | int | None = None

    def __str__(self) -> str:
        return f"{self.prefix}.{self.infix}.{self.suffix}"

    def __repr__(self) -> str:
        if self.get_value() != "Notfound":
            return f"Component(={self!s}, type={self.type_}, value={self.get_value()!r})"
        return f"Component(={self!s}, type={self.type_}, value={self.get_value()})"

    def __eq__(self, other: object) -> bool:
        parent_name = f"{self.parent.prefix}.{self.parent.infix}.{self.parent.suffix}"
        if parent_name != f"{self!s}":
            if not isinstance(other, SettingName):
                return NotImplemented
            return (self.prefix.value, self.infix.value) == (
                other.prefix.value,
                other.infix.value,
            )
        if not isinstance(other, SettingName):
            return NotImplemented
        return (self.parent.prefix.value, self.parent.infix.value) == (
            other.parent.prefix.value,
            other.parent.infix.value,
        )

    def __lt__(self, other: object) -> bool:
        parent_name = f"{self.parent.prefix}.{self.parent.infix}.{self.parent.suffix}"
        if parent_name != f"{self!s}":
            if not isinstance(other, SettingName):
                return NotImplemented
            if self.prefix.value < other.prefix.value:
                return True
            if self.prefix.value > other.prefix.value:
                return False
            if self.infix.value < other.infix.value:
                return True
            if self.infix.value > other.infix.value:
                return False
            return self.suffix.value < other.suffix.value
        if not isinstance(other, SettingName):
            return NotImplemented
        if self.parent.prefix.value < other.parent.prefix.value:
            return True
        if self.parent.prefix.value > other.parent.prefix.value:
            return False
        if self.parent.infix.value < other.parent.infix.value:
            return True
        if self.parent.infix.value > other.parent.infix.value:
            return False
        return self.parent.suffix.value < other.parent.suffix.value

    def __hash__(self):
        return hash(
            (self.prefix.value, self.infix.value, self.suffix.value)
        )

    def get_value(self) -> Any:
        uo: json | toml | yaml | mpck | None = None
        uo_exists = False
        if pathexists(
            joinpaths(current_working_dir, ".mtax", "settings.json")
        ):
            uo = json()
        if pathexists(
            joinpaths(current_working_dir, ".mtax", "settings.toml")
        ):
            uo = toml()
        if pathexists(
            joinpaths(current_working_dir, ".mtax", "settings.yaml")
        ):
            uo = yaml()
        if pathexists(
            joinpaths(current_working_dir, ".mtax", "settings.mpck")
        ):
            uo = mpck()
        if uo is None:
            uo = json()
        uo_exists = uo.exists

        mtax_obj = None
        parent_name = f"{self.parent.prefix}.{self.parent.infix}.{self.parent.suffix}"
        if uo_exists:
            try:
                if parent_name != f"{self!s}":
                    mtax_obj = uo.model(read=True)["mtax"][
                        f"{parent_name}"
                    ][f"{self!s}"]
                else:
                    mtax_obj = uo.model(read=True)["mtax"][f"{self!s}"]
            except (KeyError, JSONDecodeError):
                return "Notfound"
        if mtax_obj is not None:
            return mtax_obj
        return "Notfound"



# ----------------------------------------------------------------------
# |  Typer Helpers
# ----------------------------------------------------------------------
from .directories import date_directories
from .uo_delete import delete_useroptions
from .uo_inspect import inspect_useroptions
from .uo_init import initialize_useroptions
from .mtax import blueprints_help, commands_help, funcargs_help


__version__ = "v1.0.0"


__all__ = [
    "date_directories",
    "delete_useroptions",
    "inspect_useroptions",
    "initialize_useroptions",
    "blueprints_help",
    "commands_help",
    "funcargs_help",
    "__version__",
]



from montyalex.uo_tools._useroptions import UserOptions


# ----------------------------------------------------------------------
# |  JSON User Options Template
# ----------------------------------------------------------------------
class JsonOptions(UserOptions):
    def __init__(
        self, *, directory: str = ".mtax", filename: str = "settings"
    ) -> None:
        super().__init__(
            json=True,
            toml=False,
            yaml=False,
            mpck=False,
            directory=directory,
            filename=filename,
        )



from googletrans import LANGCODES, Translator

from montyalex.console_tools import richconsole
from montyalex.typing_tools import Any

translator = Translator()
print = richconsole.print


def lang(s: str, *, trg: str = "en", src: str = "auto") -> str:
    return translator.translate(s, dest=trg, src=src).text


def language_codes() -> dict[Any, Any]:
    return LANGCODES


language_codes_dict: dict[Any, Any] = language_codes()
langcodes: callable = language_codes
langcodes_dict: dict[Any, Any] = langcodes()


def language(
    string: str, *, target: str = "en", source: str = "auto"
) -> str:
    return lang(string, trg=target, src=source)


# ----------------------------------------------------------------------
# | Example Usage
# ----------------------------------------------------------------------
if __name__ == "__main__":
    print(lang("Hello", trg="fr"))
    print(lang("Hello", trg="it"))
    print(lang("Hello", trg="de"))



from montyalex.typer_tools import Option
from montyalex.uo_tools import json, toml, yaml, mpck
from .init_objects import (
    default_settings_template,
    empty_settings_plus_fmt,
    empty_settings_template,
)
from .initialization import initialize_schema


def initialize_useroptions(
    dirname: str = Option(".mtax", "--directory-name", "-dir"),
    filename: str = Option("settings", "--file-name", "-name"),
    *,
    overwrite: bool = Option(False, "--overwrite", "-o"),
    use_all: bool = Option(False, "--all", "-a"),
    use_synced: bool = Option(False, "--sync", "-s"),
    use_json: bool = Option(False, "--json", "-j"),
    use_toml: bool = Option(False, "--toml", "-t"),
    use_yaml: bool = Option(False, "--yaml", "-y"),
    use_mpck: bool = Option(False, "--mpck", "-m"),
):
    uo: json | toml | yaml | mpck | None = None
    if use_json:
        uo = json(directory=dirname, filename=filename)
    if use_toml:
        uo = toml(directory=dirname, filename=filename)
    if use_yaml:
        uo = yaml(directory=dirname, filename=filename)
    if use_mpck:
        uo = mpck(directory=dirname, filename=filename)
    if uo is None:
        uo = json(directory=dirname, filename=filename)

    if not use_all or use_synced:
        if overwrite:
            if isinstance(uo, json):
                uo.change("$schema", "./schema.json")
                initialize_schema(dirname)
            uo.change("mtax", default_settings_template, append=False)
        else:
            if isinstance(uo, json):
                uo.change("$schema", "./schema.json")
                initialize_schema(dirname)
            if isinstance(uo, toml):
                uo.change("mtax", empty_settings_plus_fmt)
            else:
                uo.change("mtax", default_settings_template)
    else:
        juo = json(directory=dirname, filename=filename)
        tuo = toml(directory=dirname, filename=filename)
        yuo = yaml(directory=dirname, filename=filename)
        muo = mpck(directory=dirname, filename=filename)
        if overwrite:
            juo.change("$schema", "./schema.json")
            initialize_schema(dirname)
            juo.change("mtax", default_settings_template)
            tuo.change("mtax", empty_settings_template, append=False)
            yuo.change("mtax", default_settings_template, append=False)
            muo.change("mtax", default_settings_template, append=False)
        else:
            juo.change("$schema", "./schema.json")
            initialize_schema(dirname)
            juo.change("mtax", default_settings_template)
            tuo.change("mtax", empty_settings_template)
            yuo.change("mtax", default_settings_template)
            muo.change("mtax", default_settings_template)
